"""Utilities for comparing executed queries to expectations.

These are used internally for unit testing, but can also be useful when
instrumenting code.

Version Added:
    3.4
"""

from __future__ import annotations

import operator
import re
from contextlib import contextmanager
from pprint import pformat
from typing import (Any, Callable, Dict, Iterator, List, Optional, Sequence,
                    Set, TYPE_CHECKING, Tuple, Type, TypeVar, Union, cast)

from django.db.models import F, Q, QuerySet
from django.db.models.expressions import (CombinedExpression,
                                          NegatedExpression,
                                          Subquery,
                                          Value)
from django.db.models.query import MAX_GET_RESULTS
from django.db.models.sql.query import Query as SQLQuery
from django.utils.tree import Node
from typing_extensions import Literal, NotRequired, TypeAlias, TypedDict

from djblets.db.query_catcher import catch_queries

if TYPE_CHECKING:
    from django.db.models import Model
    from django.db.models.expressions import BaseExpression

    from djblets.db.query_catcher import (CatchQueriesContext,
                                          ExecutedQueryInfo,
                                          ExecutedSubQueryInfo)

    _T = TypeVar('_T')
    _ExecutedQueryInfoT = TypeVar('_ExecutedQueryInfoT',
                                  ExecutedQueryInfo,
                                  ExecutedSubQueryInfo)


_ws_re = re.compile(r'\s+')


class QueryMismatchedAttr(TypedDict):
    """An attribute for a query that failed to meet expectations.

    Version Added:
        3.4
    """

    #: The name of the attribute.
    name: str

    #: The serialized version of the value from the executed query.
    executed_value: NotRequired[str]

    #: The serialized version of the value from the expected query.
    expected_value: NotRequired[str]

    #: The raw value from the executed query.
    raw_executed_value: NotRequired[Any]

    #: The raw value from the expected query.
    raw_expected_value: NotRequired[Any]


class QueryMismatch(TypedDict):
    """Information on a mismatched query.

    Version Added:
        3.4
    """

    #: The query that was executed.
    executed_query: SQLQuery

    #: The list of mismatched attributes in this query.
    mismatched_attrs: List[QueryMismatchedAttr]

    #: The 0-based index of the executed query that this represents.
    index: int

    #: An optional note describing this query.
    note: Optional[str]

    #: The generated SQL for the query.
    query_sql: Optional[List[str]]

    #: The results for any subquery matches.
    subqueries: Optional[CompareQueriesContext]

    #: Lines of traceback showing where this query was executed.
    traceback: Optional[List[str]]


class ExpectedQuery(TypedDict):
    """An expected query for a query assertion.

    This is used for :py:meth:`TestCase.assertQueries`.

    While this type is new, the documented keys (unless otherwise noted) were
    introduced in Djblets 2.3.1.

    Version Added:
        3.4
    """

    #: A custom note for the query, to help with query inspection.
    #:
    #: Version Added:
    #:     3.4
    __note__: NotRequired[str]

    #: A dictionary containing applied annotations.
    #:
    #: Keys are destination attribute names, and values are the annotation
    #: instances.
    #:
    #: The default is empty.
    annotations: NotRequired[Dict[str, BaseExpression]]

    #: Whether :py:meth:`django.db.models.query.QuerySet.distinct` was used.
    #:
    #: The default is ``False``.
    distinct: NotRequired[bool]

    #: A list of distinct fields.
    #:
    #: These are field names passed to
    #: :py:meth:`django.db.models.query.QuerySet.distinct`.
    #:
    #: The default is empty.
    distinct_fields: NotRequired[Tuple[str, ...]]

    #: Extra state passed in the queryset.
    #:
    #: This is state passed in calls to
    #: :py:meth:`django.db.models.query.QuerySet.extra` when using
    #: ``select`` and ``select_params``.
    #:
    #: Each key maps to a key in ``select``, and each value is a
    #: tuple containing the value in ``select`` and the corresponding
    #: value (if any) in ``select_params``.
    #:
    #: Values are normalized to collapse and strip whitespace, to
    #: help with comparison.
    #:
    #: The default is empty.
    extra: NotRequired[Dict[str, Any]]

    #: Extra order-by values.
    #:
    #: This is state passed in calls to
    #: :py:meth:`django.db.models.query.QuerySet.extra` when using
    #: ``order_by``.
    #:
    #: The default is empty.
    extra_order_by: NotRequired[Sequence[str]]

    #: Extra tables for the query.
    #:
    #: This is state passed in calls to
    #: :py:meth:`django.db.models.query.QuerySet.extra` when using ``tables``.
    #:
    #: The default is empty.
    extra_tables: NotRequired[Sequence[str]]

    #: The grouping of results.
    #:
    #: This determines whether no fields will be grouped (``None``), all
    #: fields will be grouped (``True``), or specific expressions/field
    #: names are grouped (a tuple).
    #:
    #: This is influenced by using
    #: :py:meth:`django.db.models.query.QuerySet.annotate`.
    #:
    #: The default is ``None``.
    group_by: NotRequired[Optional[Union[bool, Tuple[str, ...]]]]

    #: A mapping of joined table names to their join types.
    #:
    #: Version Added:
    #:     3.4
    join_types: NotRequired[Dict[str, str]]

    #: The value for a ``LIMIT`` in the ``SELECT``.
    #:
    #: This will generally only need to be supplied if testing a query using
    #: :py:meth:`QuerySet.exists() <django.db.models.query.QuerySet.exists>`
    #: or when slicing results.
    #:
    #: Django itself sometimes uses a default of ``None`` and sometimes a
    #: default currently of ``21`` (this exact value, and when it's used, is
    #: considered an implementation detail in Django). Both of these will
    #: match a caller-provided ``limit`` value of ``None``.
    #:
    #: The default is ``None``.
    limit: NotRequired[Optional[int]]

    #: The model backing the queryset.
    #:
    #: This represents the results that would be returned or altered by the
    #: query.
    model: NotRequired[Type[Model]]

    #: The number of tables joined.
    #:
    #: The default is 0.
    num_joins: NotRequired[int]

    #: The value for an ``OFFSET`` in the ``SELECT``.
    #:
    #: The default is 0.
    offset: NotRequired[int]

    #: The specific fields being fetched, or ``None`` if fetching all fields.
    #:
    #: The default is ``None``.
    only_fields: NotRequired[Set[str]]

    #: The ordering criteria.
    #:
    #: The default is empty.
    order_by: NotRequired[Tuple[str, ...]]

    #: Whether this is a select-for-update operation.
    #:
    #: The default is ``False``.
    select_for_update: NotRequired[bool]

    #: The table names involved in a select-related operation.
    #:
    #: These are the names passed to
    #: :py:meth:`django.db.models.query.QuerySet.select_related`. If called
    #: without any parameters, this would be ``True``.
    select_related: NotRequired[Union[Literal[True], Set[str]]]

    #: Information on subqueries within this query.
    #:
    #: Version Added:
    #:     3.4
    subqueries: NotRequired[Optional[ExpectedQueries]]

    #: Whether this is considered a subquery of another query.
    #:
    #: The default is ``False``.
    subquery: NotRequired[bool]

    #: The tables involved in the query.
    #:
    #: The default is the model's table name.
    tables: NotRequired[Set[str]]

    #: The query type.
    #:
    #: This would be one of ``DELETE``, ``INSERT``, ``SELECT``, or ``UPDATE``.
    #:
    #: The default is ``SELECT``.
    type: NotRequired[str]

    #: A list of specified fields to return as values.
    #:
    #: These are values passed in
    #: :py:meth:`~django.db.models.query.QuerySet.values` or
    #: :py:meth:`~django.db.models.query.QuerySet.values_list`.
    values_select: NotRequired[Tuple[str, ...]]

    #: The query expression objects used to represent the filter on the query.
    #:
    #: These are normalized to filter out empty or unnecessarily-nested
    #: queries, to ease comparison.
    where: NotRequired[Q]


#: A type representing list of expected queries.
#:
#: Note:
#:     This type allows either a :py:meth:`ExpectedQuery` or a plain
#:     dictionary of string-based keys. This is because with pyright (as of
#:     1.1.332), the type of the right-hand-side of a ``+=`` operation cannot
#:     be inferred. That makes it difficult to dynamically build expected
#:     queries.
#:
#:     Until/unless the situation improves, we aim for an
#:     :py:meth:`ExpectedQuery` but allow a plain dictionary.
#:
#: Version Added:
#:     3.4
ExpectedQueries: TypeAlias = List[Union[ExpectedQuery,
                                        Dict[str, Any]]]


class CompareQueriesContext(TypedDict):
    """Context for compared query results.

    This is provided and populated when using :py:func:`compare_queries`.

    Version Added:
        3.4
    """

    #: Whether there are any mismatches found.
    has_mismatches: bool

    #: The number of queries that were executed.
    num_executed_queries: int

    #: The number of queries that were expected.
    num_expected_queries: int

    #: Whether the number of executed and expected queries are different.
    query_count_mismatch: bool

    #: The list of query expectation failures.
    query_mismatches: List[QueryMismatch]

    #: An internal list of unchecked mismatched attributes.
    #:
    #: This is used internally and may be removed in a future version
    #: without a deprecation period.
    _unchecked_mismatched_attrs: Set[str]


class _CheckQueryResult(TypedDict):
    """Results for a _check_query call.

    Version Added:
        3.4
    """

    #: A list of mismatched attributes for this query.
    #:
    #: This will be empty if there are no mismatches.
    mismatched_attrs: List[QueryMismatchedAttr]

    #: The comparison context for any subqueries.
    #:
    #: This will be ``None`` if not checking for subqueries.
    subqueries_compare_ctx: Optional[CompareQueriesContext]

    #: An internal list of unchecked mismatched attributes.
    unchecked_mismatched_attrs: Set[str]


@contextmanager
def compare_queries(
    queries: Sequence[Union[ExpectedQuery,
                            Dict[str, Any]]],
    *,
    _check_join_types: bool = True,
    _check_subqueries: bool = True,
) -> Iterator[CompareQueriesContext]:
    """Assert the number and complexity of queries.

    This provides advanced checking of queries, allowing the caller to
    match filtering, JOINs, ordering, selected fields, and more.

    This takes a list of dictionaries with query information. Each
    contains the keys in :py:class:`ExpectedQuery`.

    Version Added:
        3.4

    Args:
        queries (list of ExpectedQuery):
            The list of query dictionaries to compare executed queries
            against.

        _check_join_types (bool, optional):
            Whether to check join types.

            This is internal for compatibility with the old behavior for
            :py:meth:`TestCase.assertQueries()
            <djblets.testing.testcases.assertQueries>` and will be removed in a
            future release without a deprecation period.

        _check_subqueries (bool, optional):
            Whether to check subqueries.

            This is internal for compatibility with the old behavior for
            :py:meth:`TestCase.assertQueries()
            <djblets.testing.testcases.assertQueries>` and will be removed in a
            future release without a deprecation period.

    Context:
        dict:
        The context for compared queries.

        This will only be populated after the context manager has finished.
        See :py:class:`CompareQueriesContext` for details.
    """
    assert isinstance(queries, list)

    compare_ctx = CompareQueriesContext(
        has_mismatches=False,
        num_executed_queries=0,
        num_expected_queries=0,
        query_count_mismatch=False,
        query_mismatches=[],
        _unchecked_mismatched_attrs=set())

    with catch_queries(_check_subqueries=_check_subqueries) as catch_ctx:
        yield compare_ctx

    compare_ctx.update(_check_queries(
        expected_queries=queries,
        executed_queries=catch_ctx.executed_queries,
        catch_ctx=catch_ctx,
        _check_join_types=_check_join_types,
        _check_subqueries=_check_subqueries))


def _check_queries(
    *,
    expected_queries: List[Union[ExpectedQuery,
                                 Dict[str, Any]]],
    executed_queries: Sequence[_ExecutedQueryInfoT],
    catch_ctx: CatchQueriesContext,
    _check_join_types: bool = True,
    _check_subqueries: bool = True,
) -> CompareQueriesContext:
    """Check database queries for expected output.

    This takes a list of executed queries captured using
    :py:func:`~djblets.db.query_catcher.catch_queries` and compare them against
    a list of expected queries. The results are a list of mismatches found
    between the executed and expeceted queries.

    Version Added:
        3.4

    Args:
        expected_queries (list of dict):
            The list of expected queries.

        executed_queries (list of dict):
            The list of executed queries.

        catch_ctx (djblets.db.query_catcher.CatchQueriesContext):
            The context for caught queries.

        _check_join_types (bool, optional):
            Whether to check join types.

            This is internal for compatibility with the old behavior for
            :py:meth:`TestCase.assertQueries()
            <djblets.testing.testcases.assertQueries>` and will be removed in a
            future release without a deprecation period.

        _check_subqueries (bool, optional):
            Whether to check subqueries.

            This is internal for compatibility with the old behavior for
            :py:meth:`TestCase.assertQueries()
            <djblets.testing.testcases.assertQueries>` and will be removed in a
            future release without a deprecation period.

    Returns:
        dict:
        The context for compared queries.

        See :py:class:`CompareQueriesContext`.
    """
    # Now we can compare numbers.
    num_expected_queries = len(expected_queries)
    num_executed_queries = len(executed_queries)
    query_count_mismatch: bool = False
    unchecked_mismatched_attrs = set()

    # Make sure we received the expected number of queries.
    if num_expected_queries != num_executed_queries:
        query_count_mismatch = True

        if num_expected_queries < num_executed_queries:
            expected_queries += [
                {}
                for _i in range(num_executed_queries - num_expected_queries)
            ]
        elif num_expected_queries > num_executed_queries:
            expected_queries = expected_queries[:num_executed_queries]

    # Go through each matching Query and compare state.
    query_mismatches: List[QueryMismatch] = []
    queries_iter = enumerate(zip(cast(Sequence[ExpectedQuery],
                                      expected_queries),
                                 executed_queries))

    for i, (query_info, executed_query_info) in queries_iter:
        executed_query = executed_query_info['query']
        result = _check_query(
            executed_query=executed_query,
            executed_query_info=executed_query_info,
            expected_query_info=query_info,
            check_join_types=_check_join_types,
            check_subqueries=_check_subqueries,
            catch_ctx=catch_ctx)
        mismatched_attrs = result['mismatched_attrs']
        subqueries_compare_ctx = result.get('subqueries_compare_ctx')
        unchecked_mismatched_attrs.update(result['unchecked_mismatched_attrs'])

        if (mismatched_attrs or
            (_check_subqueries and
             subqueries_compare_ctx is not None and
             subqueries_compare_ctx['has_mismatches'])):
            query_mismatches.append({
                'executed_query': executed_query,
                'index': i,
                'mismatched_attrs': mismatched_attrs,
                'note': query_info.get('__note__'),
                'query_sql': cast(Optional[List[str]],
                                  executed_query_info.get('sql')),
                'subqueries': subqueries_compare_ctx,
                'traceback': cast(Optional[List[str]],
                                  executed_query_info.get('traceback')),
            })

    # Update the context for the caller to examine.
    return {
        'has_mismatches': query_count_mismatch or (len(query_mismatches) > 0),
        'num_executed_queries': num_executed_queries,
        'num_expected_queries': num_expected_queries,
        'query_count_mismatch': query_count_mismatch,
        'query_mismatches': query_mismatches,
        '_unchecked_mismatched_attrs': unchecked_mismatched_attrs,
    }


def _check_expectation(
    name: str,
    *,
    mismatched_attrs: List[QueryMismatchedAttr],
    expected_value: _T,
    executed_value: _T,
    match_func: Callable[[_T, _T], bool] = operator.__eq__,
    format_expected_value_func: Callable[[_T], str] = pformat,
    format_executed_value_func: Callable[[_T], str] = pformat,
) -> None:
    """Check the expectation for a query attribute.

    This will compare the query attribute values between the expected and
    executed queries, storing a mismatch in ``mismatched_attrs`` containing
    raw and human-readable versions of the values.

    Version Added:
        3.4

    Args:
        name (str):
            The name of the attribute to check.

        mismatched_attrs (list of dict):
            The list of mismatched attributes to apapend to if there's a
            mismatch.

        expected_value (object):
            The expected value for the attribute.

        executed_value (object):
            The executed value for the attribute.

        match_func (callable, optional):
            The function to call to compare the values.

            This defaults to an equality check.

        format_expected_value_func (callable, optional):
            The function to format the expected value for the human-readable
            representation.

            This defaults to :py:func:`pprint.pformat`.

        format_executed_value_func (callable, optional):
            The function to format the executed value for the human-readable
            representation.

            This defaults to :py:func:`pprint.pformat`.
    """
    if not match_func(expected_value, executed_value):
        mismatched_attrs.append({
            'name': name,
            'executed_value': format_executed_value_func(executed_value),
            'expected_value': format_expected_value_func(expected_value),
            'raw_executed_value': executed_value,
            'raw_expected_value': expected_value,
        })


def _check_query(
    *,
    executed_query: SQLQuery,
    executed_query_info: _ExecutedQueryInfoT,
    expected_query_info: ExpectedQuery,
    check_join_types: bool,
    check_subqueries: bool,
    catch_ctx: CatchQueriesContext,
) -> _CheckQueryResult:
    """Check a query for expectations.

    This will compare attributes on an executed query to an expected query
    and determine if there are any mismatches.

    Version Added:
        3.4

    Args:
        executed_query (django.db.models.sql.query.Query):
            The executed query to compare.

        executed_query_info (djblets.db.query_catcher.ExecutedQueryInfo):
            Information on the executed query to use for comparison.

        expected_query_info (dict):
            Information on the expectations for the query from the caller.

        check_join_types (bool):
            Whether to check for join types.

            This is internal for compatibility with the old behavior for
            :py:meth:`TestCase.assertQueries()
            <djblets.testing.testcases.assertQueries>` and will be removed in a
            future release without a deprecation period.

        check_subqueries (bool):
            Whether to check for subqueries.

            This is internal for compatibility with the old behavior for
            :py:meth:`TestCase.assertQueries()
            <djblets.testing.testcases.assertQueries>` and will be removed in a
            future release without a deprecation period.

        catch_ctx (djblets.db.query_catcher.CatchQueriesContext):
            The context for caught queries.

    Returns:
        dict:
        A dictionary of results.

        See :py:class:`_CheckQueryResult` for details.
    """
    mismatched_attrs: List[QueryMismatchedAttr] = []
    unchecked_mismatched_attrs: Set[str] = set()

    executed_model = executed_query.model
    assert executed_model is not None

    expected_model = expected_query_info.get('model')

    executed_table_name = executed_model._meta.db_table
    executed_reffed_tables: Set[str] = {
        _table_name
        for _table_name in executed_query.alias_map.keys()
        if executed_query.alias_refcount.get(_table_name, 0) > 0
    }

    # Check 'model'.
    _check_expectation(
        'model',
        mismatched_attrs=mismatched_attrs,
        expected_value=expected_model,
        executed_value=executed_model,
        match_func=operator.is_)

    # Check the query type.
    _check_expectation(
        'type',
        mismatched_attrs=mismatched_attrs,
        expected_value=expected_query_info.get('type', 'SELECT'),
        executed_value=executed_query_info['type'].value)

    # Check the fields that are easy to compare.
    for key, default in (('annotations', {}),
                         ('distinct', False),
                         ('distinct_fields', ()),
                         ('extra_order_by', ()),
                         ('extra_tables', ()),
                         ('group_by', None),
                         ('order_by', ()),
                         ('select_for_update', False),
                         ('subquery', False),
                         ('values_select', ())):
        _check_expectation(
            key,
            mismatched_attrs=mismatched_attrs,
            expected_value=expected_query_info.get(key, default),
            executed_value=getattr(executed_query, key))

    # Check 'extra'.
    _check_expectation(
        'extra',
        mismatched_attrs=mismatched_attrs,
        expected_value={
            _key: (_ws_re.sub(' ', _value[0]).strip(), _value[1])
            for _key, _value in expected_query_info.get('extra',
                                                        {}).items()
        },
        executed_value={
            _key: (_ws_re.sub(' ', _value[0]).strip(), _value[1])
            for _key, _value in executed_query.extra.items()
        })

    # Check 'offset'.
    _check_expectation(
        'offset',
        mismatched_attrs=mismatched_attrs,
        expected_value=expected_query_info.get('offset', 0),
        executed_value=executed_query.low_mark)

    # Check 'limit'.
    _check_expectation(
        'limit',
        mismatched_attrs=mismatched_attrs,
        expected_value=expected_query_info.get('limit'),
        executed_value=executed_query.high_mark,
        match_func=lambda expected_value, executed_value: (
            expected_value == executed_value or
            (expected_value is None and
             executed_value == MAX_GET_RESULTS)))

    # Check 'num_joins'.
    if executed_query.alias_map:
        # Consider all tables that have been JOINed, ignoring the
        # main table.
        executed_num_joins = len(executed_reffed_tables -
                                 {executed_table_name})
    else:
        executed_num_joins = 0

    _check_expectation(
        'num_joins',
        mismatched_attrs=mismatched_attrs,
        expected_value=expected_query_info.get('num_joins', 0),
        executed_value=executed_num_joins)

    # Check 'tables'.
    tables_default: Set[str]

    if expected_model is None:
        tables_default = set()
    else:
        tables_default = {expected_model._meta.db_table}

    if executed_query.alias_refcount:
        executed_tables = \
            executed_reffed_tables | {executed_table_name}
    else:
        executed_tables = {executed_table_name}

    _check_expectation(
        'tables',
        mismatched_attrs=mismatched_attrs,
        expected_value=set(expected_query_info.get('tables',
                                                   tables_default)),
        executed_value=executed_tables,
        format_expected_value_func=_format_set,
        format_executed_value_func=_format_set)

    # Check 'join_types'.
    executed_join_types = {
        _table_name: _value.join_type
        for _table_name, _value in executed_query.alias_map.items()
        if (_table_name != executed_table_name and
            executed_query.alias_refcount.get(_table_name))
    }

    if check_join_types:
        _check_expectation(
            'join_types',
            mismatched_attrs=mismatched_attrs,
            expected_value=expected_query_info.get('join_types', {}),
            executed_value=executed_join_types)
    elif executed_join_types and 'join_types' not in expected_query_info:
        unchecked_mismatched_attrs.add('join_types')

    # Check 'only_fields'.
    _check_expectation(
        'only_fields',
        mismatched_attrs=mismatched_attrs,
        expected_value=set(expected_query_info.get('only_fields',
                                                   set())),
        executed_value=set(executed_query.deferred_loading[0]),
        format_expected_value_func=_format_set,
        format_executed_value_func=_format_set)

    # Check 'select_related'.
    executed_select_related_raw = executed_query.select_related
    executed_select_related: Union[Literal[True], Set[str]]

    if executed_select_related_raw is True:
        executed_select_related = True
    elif isinstance(executed_select_related_raw, dict):
        executed_select_related = \
            set(executed_select_related_raw.keys())
    else:
        executed_select_related = set()

    expected_select_related_raw = \
        expected_query_info.get('select_related')
    expected_select_related: Union[Literal[True], Set[str]]

    if expected_select_related_raw is True:
        expected_select_related = True
    elif expected_select_related_raw:
        expected_select_related = set(expected_select_related_raw)
    else:
        expected_select_related = set()

    _check_expectation(
        'select_related',
        mismatched_attrs=mismatched_attrs,
        expected_value=expected_select_related,
        executed_value=executed_select_related)

    # Check 'where'. Normalize the Q object used to filter.
    _check_expectation(
        'where',
        mismatched_attrs=mismatched_attrs,
        expected_value=_normalize_q(
            expected_query_info.get('where', Q()),
            normalize_subqueries=False),
        executed_value=_normalize_q(
            catch_ctx.queries_to_qs.get(executed_query, Q()),
            normalize_subqueries=check_subqueries),
        format_expected_value_func=(
            lambda q: _format_node(q, catch_ctx=catch_ctx)),
        format_executed_value_func=(
            lambda q: _format_node(q, catch_ctx=catch_ctx)))

    # Check if this is an aggregate query.
    subqueries_compare_ctx: Optional[CompareQueriesContext] = None

    # Check any subqueries.
    if check_subqueries:
        subqueries_compare_ctx = _check_queries(
            expected_queries=expected_query_info.get('subqueries') or [],
            executed_queries=executed_query_info['subqueries'],
            _check_join_types=check_join_types,
            catch_ctx=catch_ctx)

    return {
        'mismatched_attrs': mismatched_attrs,
        'subqueries_compare_ctx': subqueries_compare_ctx,
        'unchecked_mismatched_attrs': unchecked_mismatched_attrs,
    }


def _build_subquery_placeholder(
    *,
    subquery: Union[Q, QuerySet, Subquery],
    subqueries: List[Any],
) -> Union[Node, Q, tuple]:
    """Normalize a subquery to a placeholder.

    This will convert a subquery or an expression containing subquery value to
    a placeholder Q object in the form of::

        Q(__SubqueryType__subquery__=index)

    This allows for consistent comparison between subqueries, with detailed
    comparisons happening in expected subquery lists.

    Args:
        subquery (django.db.models.Q or
                  django.db.models.QuerySet or
                  django.db.models.expressions.Subquery):
            The subquery to normalize.

        subqueries (list):
            A list for tracking normalized subqueries.

    Returns:
        django.db.models.Q:
        The resulting normalized Q object.
    """
    subqueries.append(subquery)
    subquery_type = type(subquery).__name__
    norm_subquery = (f'__{subquery_type}__subquery__', len(subqueries))

    if getattr(subquery, 'negated', False):
        return ~Q(norm_subquery)
    else:
        return norm_subquery


def _normalize_q(
    q: Q,
    *,
    subqueries: Optional[List[Any]] = None,
    normalize_subqueries: bool = True,
) -> Q:
    """Normalize a Q object for comparison.

    This will filter out any empty Q objects, which are normally filtered
    out during query execution. This helps to compare Q objects without
    worrying whether a code path includes an empty Q object or omits it
    entirely.

    By default, subqueries will be normalized to
    ``Q(__SubqueryName__subquery__=index)`` objects for comparison purposes.

    Version Added:
        3.4

    Args:
        q (django.db.models.Q):
            The Q object to normalize.

        subqueries (list, optional):
            A list of subqueries found, for tracking purposes.

            This is considered internal for recursive calls, and does not
            need to be supplied.

        normalize_subqueries (bool, optional):
            Whether to normalize subqueries.

    Returns:
        django.db.models.Q:
        The resulting Q object.
    """
    if subqueries is None:
        subqueries = []

    children: List[Any] = []

    # Negated expressions are special. We need to extract an expressin out of
    # them.
    is_negated_epxression = False and isinstance(q, NegatedExpression)

    if is_negated_epxression:
        q_children = [q.expression]
    else:
        q_children = getattr(q, 'children', [])

    # Go through all the children and see if we can collapse or transform any
    # of them for comparison.
    for child in q_children:
        if isinstance(child, Q):
            # Remove any empty children.
            if not child:
                continue

            # Normalize the child, recursively.
            child = _normalize_q(child,
                                 subqueries=subqueries,
                                 normalize_subqueries=normalize_subqueries)
            grandchildren = child.children

            if (len(grandchildren) == 1 and
                isinstance(grandchildren[0], tuple)):
                # This is a Q() with only a single value in it. We may be
                # able to collapse or record a subquery.
                grandchild = grandchildren[0]

                if (normalize_subqueries and
                    isinstance(grandchild[1], QuerySet)):
                    child = _build_subquery_placeholder(
                        subquery=child,
                        subqueries=subqueries)
                elif not child.negated:
                    # Since we're not negating, we can collapse the child.
                    child = grandchild
        elif isinstance(child, NegatedExpression):
            child = ~Q(_build_subquery_placeholder(
                subquery=child.expression,
                subqueries=subqueries))
        elif normalize_subqueries and isinstance(child, Subquery):
            # Subqueries are formatted as Q(__SubqueryType__subquery__=index),
            # since we can't actually compare true subqueries. Instead, this
            # will reference a subquery in the subqueries dictionaries in the
            # compare results.
            child = _build_subquery_placeholder(
                subquery=child,
                subqueries=subqueries)

        children.append(child)

    new_q = q

    if is_negated_epxression:
        assert len(children) == 1

        return children[0]

    if len(children) == 1 and isinstance(children[0], Q):
        # There's only one single nested Q, meaning we don't have multiple Qs
        # or any tuples normalized above. We can now determine if we can safely
        # swap in the child and its state.
        child = children[0]
        q_negated = q.negated

        if not q_negated:
            # The current Q object doesn't have a negated property, and it only
            # has one child, so we can optimize it out.
            # collapse down
            new_q = child

        if not child.negated:
            # The child doesn't negate its value, so we can safely use the
            # original Q object's negation state. If the child had negated its
            # value, we'd want to preserve that.
            new_q.negated = q_negated

    if new_q is q:
        new_q.children = children

    return new_q


def _format_query_value(
    value: Any,
) -> str:
    """Format a value in a query.

    This will format a combined expression (such as an F expression) or any
    other value in a query.

    Version Added:
        3.4

    Args:
        value (object):
            The value to format.

    Returns:
        str:
        The formatted value.
    """
    if isinstance(value, CombinedExpression):
        return '%s %s %s' % (_format_query_value(value.lhs),
                             value.connector,
                             _format_query_value(value.rhs))
    elif isinstance(value, F):
        return "%s('%s')" % (type(value).__name__, value.name)
    elif isinstance(value, Value):
        return str(value.value)
    elif isinstance(value, set):
        if value:
            return '{%s}' % ', '.join(
                repr(_item)
                for _item in sorted(value)
            )
        else:
            return 'set()'
    else:
        return repr(value)


def _format_query_part(
    part: Any,
    *,
    indent: str,
    prefix: str = '',
    in_negated: bool = False,
    catch_ctx: CatchQueriesContext,
) -> str:
    """Format part of a query.

    This will format a Node, Query, tuple, or other expression within a query
    tree for display.

    Version Added:
        3.4

    Args:
        query (object):
            The query part object in the tree to format.

        indent (int, optional):
            The indentation level for the formatted output.

        prefix (str, optional):
            The optional prefix for any outputted lines.

        catch_ctx (djblets.db.query_catcher.CatchQueriesContext):
            The context for caught queries.

    Returns:
        str:
        The formatted part.
    """
    if isinstance(part, Node):
        # This is a node (e.g., a Q() object or similar). We can format it
        # as usual.
        return _format_node(part,
                            catch_ctx=catch_ctx,
                            indent=indent,
                            prefix=prefix)
    elif isinstance(part, SQLQuery):
        # This is another query. We need to grab the associated Q() and
        # format that.
        return _format_node(catch_ctx.queries_to_qs[part],
                            catch_ctx=catch_ctx,
                            indent=indent,
                            prefix=prefix)
    elif isinstance(part, tuple):
        # For tuples, just output a `Q()` normally to avoid wrapping inside
        # of another Q().
        key, value = part
        formatted_value = _format_query_value(value)
        part = f'Q({key}={formatted_value})'
    else:
        part = _format_query_value(part)

    # Format either the tuple or whatever else we found as a string
    # representation.
    return f'{indent}{prefix}{part}'


def _format_query_children(
    query: Union[BaseExpression, Node],
    *,
    indent: str,
    op: str = '',
    in_negated: bool = False,
    catch_ctx: CatchQueriesContext,
) -> str:
    """Format children of a node or expression.

    This will iterate through all children, formatting them with an optional
    prefixed operator (for AND or OR queries) at the provided indentation
    level.

    Version Added:
        3.4

    Args:
        node (django.utils.tree.Node or django.db.expressions.BaseExpression):
            The node or expression containing children to format.

        indent (int, optional):
            The indentation level for the formatted output.

        op (str, optional):
            The optional operator to prepend to all lines after the first.

        catch_ctx (djblets.db.query_catcher.CatchQueriesContext):
            The context for caught queries.

    Returns:
        str:
        The formatted Node object.
    """
    children = getattr(query, 'children', [])
    init_parts: List[str]

    if op and len(children) > 1:
        # The first line won't contain an operator, but does need padding for
        # that operator. We'll add this and then process the rest of the
        # children with the operator prefixed to each line.
        prefix = f'{op} '
        init_parts = [
            _format_query_part(children[0],
                               indent=indent,
                               prefix=' ' * len(prefix),
                               catch_ctx=catch_ctx,
                               in_negated=in_negated),
        ]

        children = children[1:]
    else:
        prefix = ''
        init_parts = []

    return '\n'.join(init_parts + [
        _format_query_part(child,
                           indent=indent,
                           prefix=prefix,
                           in_negated=in_negated,
                           catch_ctx=catch_ctx)
        for child in children
    ])


def _format_node(
    node: Node,
    *,
    indent: str = '',
    prefix: str = '',
    catch_ctx: CatchQueriesContext,
) -> str:
    """Format a Node object for easier comparison.

    This will format a normalized Node object and all children in a nested tree
    form, helping to see the structure of the query and compare differences.

    This expects all nodes to be normalized via :py:func:`_normalize_q`.

    Version Added:
        3.4

    Args:
        node (django.utils.tree.Node):
            The node object to format.

        indent (int, optional):
            The indentation level for the formatted output.

        prefix (str, optional):
            The optional prefix for any outputted lines.

        catch_ctx (djblets.db.query_catcher.CatchQueriesContext):
            The context for caught queries.

    Returns:
        str:
        The formatted Node object.
    """
    name = type(node).__name__
    children_str: str = ''

    if node.negated:
        name = f'~{name}'

    if len(node.children) == 1:
        # We may need to collapse this down for display, as Q() objects nest
        # internally (even after normalization). If it's a tuple or a
        # single-child Q(), then we can safely do this.
        child = node.children[0]

        if isinstance(child, tuple):
            # For a tuple, we'll just represent the key/value in the
            # generated Q(...) below.
            key, value = child
            formatted_value = _format_query_value(value)
            children_str = f'{key}={formatted_value}'
        elif isinstance(child, Q) and len(child.children) > 1:
            # For a Q(), we'll collapse down by formatting using the
            # child. This is done since negated Qs have an AND Q inside,
            # and that's the one we want to show.
            node = child

    if node.connector == Q.AND:
        op = '&'
    elif node.connector == Q.OR:
        op = '|'
    else:
        op = ''

    if not children_str:
        prefix_indent = ' ' * len(prefix)
        children_str = _format_query_children(
            node,
            indent=f'{indent}{prefix_indent} ',
            op=op,
            catch_ctx=catch_ctx)

        if children_str:
            children_str = f'\n{children_str}'

    return f'{indent}{prefix}{name}({children_str})'


def _format_set(
    values: Set[Any],
) -> str:
    """Format a set with sorted values.

    :py:mod:`pprint` doesn't sort sets unless they span multiple lines of
    output. This function takes the set, turns it into a sorted list, and
    then re-represents it as sorted set.

    Args:
        values (set):
            The set to format.

    Returns:
        str:
        The formatted set.
    """
    return '{%s}' % (pformat(sorted(values))[1:-1])



from django.urls import (NoReverseMatch,
                         clear_url_caches,
                         include,
                         path,
                         reverse)
from django.views.decorators.cache import never_cache
from kgb import SpyAgency

from djblets.testing.testcases import TestCase
from djblets.urls.patterns import never_cache_patterns
from djblets.urls.resolvers import DynamicURLResolver


def dummy_view(request):
    pass


class URLPatternsTests(SpyAgency, TestCase):
    """Unit tests for djblets.urls.patterns."""

    def test_never_cache_patterns(self):
        """Testing never_cache_patterns"""
        self.spy_on(never_cache)

        urlpatterns = never_cache_patterns(
            path('a/', dummy_view),
            path('b/', dummy_view),
        )

        self.assertEqual(len(never_cache.spy.calls), 2)
        self.assertEqual(len(urlpatterns), 2)

        self.assertTrue(
            never_cache.spy.calls[0].returned(urlpatterns[0].callback))
        self.assertTrue(
            never_cache.spy.calls[1].returned(urlpatterns[1].callback))


class URLResolverTests(TestCase):
    def tearDown(self):
        super(URLResolverTests, self).tearDown()
        clear_url_caches()

    def test_dynamic_url_resolver(self):
        """Testing DynamicURLResolver"""
        def dummy_view(self):
            pass

        dynamic_urls = DynamicURLResolver()
        root_urlconf = (
            path('root/', include([dynamic_urls])),
            path('foo/', dummy_view, name='foo'),
        )

        with self.settings(ROOT_URLCONF=root_urlconf):
            clear_url_caches()

            new_patterns = [
                path('bar/', dummy_view, name='bar'),
                path('baz/', dummy_view, name='baz'),
            ]

            # The new patterns shouldn't reverse, just the original "foo".
            reverse('foo')
            self.assertRaises(NoReverseMatch, reverse, 'bar')
            self.assertRaises(NoReverseMatch, reverse, 'baz')

            # Add the new patterns. Now reversing should work.
            dynamic_urls.add_patterns(new_patterns)

            reverse('foo')
            reverse('bar')
            reverse('baz')

            # Get rid of the patterns again. We should be back in the original
            # state.
            dynamic_urls.remove_patterns(new_patterns)

            reverse('foo')
            self.assertRaises(NoReverseMatch, reverse, 'bar')
            self.assertRaises(NoReverseMatch, reverse, 'baz')



"""Testing utilities for mail-related unit tests."""

from __future__ import annotations

from typing import Dict, List, TYPE_CHECKING, Union

import dns.message
import dns.rdataclass
import dns.rdatatype
import dns.resolver
from django.core.cache import cache
from django.utils.encoding import force_bytes
from dns.rdtypes.ANY.TXT import TXT
from kgb import SpyAgency

if TYPE_CHECKING:
    from djblets.testing.testcases import TestCase as MixinParentClass
else:
    MixinParentClass = object


class DmarcDnsTestsMixin(MixinParentClass):
    """Mixin to help with e-mail tests that need to perform DMARC lookups.

    This mixin makes it easy for unit tests to fake DMARC results, in order
    to prevent the need from looking up real data from real records.

    Unit tests can provide a mapping from domain names to TXT record strings
    in the :py:attr:`dmarc_txt_records` dictionary. When a DMARC lookup is
    performed, this dictionary will be used for the lookup.

    Note that this mixin will also clear the memory cache before each test
    run.
    """

    #: A dictionary of domain names to DMARC TXT record data.
    #:
    #: Type:
    #:     dict
    dmarc_txt_records: Dict[str, Union[bytes, str, list[Union[bytes, str]]]]

    def setUp(self) -> None:
        self.dmarc_txt_records = {}

        self._dmarc_spy_agency = SpyAgency()
        self._dmarc_spy_agency.spy_on(dns.resolver.resolve,
                                      call_fake=self._dns_query)
        cache.clear()

        # This has to happen after we clear the cache. Some other test cases
        # (such as webapi or siteconfig tests) rely on cached data that gets
        # set up in their setUp methods.
        super().setUp()

    def tearDown(self) -> None:
        super().tearDown()

        self._dmarc_spy_agency.unspy_all()

    def _dns_query(
        self,
        qname: str,
        *args,
        **kwargs,
    ) -> List:
        """Return a fake answer for a DNS query.

        This will return either a TXT record or a NXDOMAIN error based on the
        presence of a key in :py:attr:`dmarc_txt_records`.

        Note that technically this should be returning a
        :py:class:`dns.resolver.Answer`, but that's fairly complex to set up,
        and our usage turns this into a list of results anyway.

        Args:
            qname (str):
                The domain being queried.

            *args (tuple, unused):
                Unused positional arguments.

            **kwargs (tuple, unused):
                Unused keyword arguments.

        Returns:
            list:
            The list of resulting records.

        Raises:
            dns.resolve.NXDOMAIN:
                The domain did not have a pre-populated result.
        """
        try:
            strings: list[Union[bytes, str]]
            value = self.dmarc_txt_records[qname]

            if isinstance(value, list):
                strings = value
            else:
                strings = [value]

            return [TXT(1, 16, [
                force_bytes(s)
                for s in strings
            ])]
        except KeyError:
            raise dns.resolver.NXDOMAIN



#!/usr/bin/env python3
"""Write node.js dependencies to djblets/dependencies.py.

Version Added:
    4.0
"""

import json
import os
from typing import Dict, TextIO


MARKER_START = '# Auto-generated Node.js dependencies {\n'
MARKER_END = '# } Auto-generated Node.js dependencies\n'


LINT_DEP_NAMES = {
    'eslint',
    '@beanbag/eslint-plugin',
}


def _write_deps(
    *,
    fp: TextIO,
    doc: str,
    name: str,
    deps: Dict[str, str],
) -> None:
    """Write dependencies to the file.

    This will write the Python code to list each dependency and the
    matching version.

    Args:
        fp (io.TextIO):
            The file to write to.

        doc (str):
            The doc comment contents.

        name (str):
            The name of the variable to write to.

        deps (dict):
            The dependencies to write.
    """
    fp.write(
        '#: %(doc)s\n'
        '%(name)s: Dict[str, str] = {\n'
        '%(deps)s\n'
        '}\n'
        '\n'
        % {
            'doc': doc,
            'name': name,
            'deps': '\n'.join(
                f"    '{dep_name}': '{dep_ver}',"
                for dep_name, dep_ver in deps.items()
            ),
        })


def main() -> None:
    """Embed package.json into djblets/dependencies.py."""
    scripts_dir = os.path.abspath(os.path.dirname(__file__))
    top_dir = os.path.abspath(os.path.join(scripts_dir, '..', '..'))
    djblets_dir = os.path.join(top_dir, 'djblets')
    package_json_path = os.path.join(djblets_dir, 'package.json')
    deps_py_path = os.path.join(djblets_dir, 'dependencies.py')

    # Load the dependencies and organize them.
    with open(package_json_path, 'r') as fp:
        package_json = json.load(fp)

    frontend_deps: Dict[str, str] = {}
    lint_deps: Dict[str, str] = {}

    for dep_name, dep_ver in sorted(package_json['dependencies'].items()):
        if dep_name in LINT_DEP_NAMES:
            lint_deps[dep_name] = dep_ver
        else:
            frontend_deps[dep_name] = dep_ver

    # Parse out the existing dependencies.py and grab everything outside the
    # markers.
    new_lines_pre: str = ''
    new_lines_post: str = ''

    with open(deps_py_path, 'r') as fp:
        data = fp.read()

        i = data.find(MARKER_START)
        assert i != -1

        j = data.find(MARKER_END, i)
        assert j != -1

        new_lines_pre = data[:i]
        new_lines_post = data[j + len(MARKER_END) + 1:]

    # Write out the new dependencies.py.
    with open(deps_py_path, 'w') as fp:
        fp.write(new_lines_pre)
        fp.write(f'{MARKER_START}\n\n')

        _write_deps(
            fp=fp,
            doc='Dependencies required for static media building.',
            name='frontend_buildkit_npm_dependencies',
            deps=frontend_deps)

        _write_deps(
            fp=fp,
            doc='Dependencies required for static media linting.',
            name='lint_npm_dependencies',
            deps=lint_deps)

        fp.write(f'\n{MARKER_END}\n')
        fp.write(new_lines_post)


if __name__ == '__main__':
    main()



"""Unit tests for djblets.db.fields.comma_separated_values_field."""

from __future__ import annotations

from django.db import models

from djblets.db.fields import CommaSeparatedValuesField
from djblets.testing.testcases import TestCase, TestModelsLoaderMixin


class TestModel(models.Model):
    """Test model for the CommaSeparatedValuesField."""

    field = CommaSeparatedValuesField(
        max_length=512,
        blank=True)


class CommaSeparatedValuesFieldTests(TestModelsLoaderMixin, TestCase):
    """Tests for djblets.db.fields.comma_separated_values_field."""

    tests_app = 'djblets.db.tests'

    def test_empty_field(self) -> None:
        """Testing CommaSeparatedValuesField with empty initial value."""
        obj = TestModel()
        self.assertEqual(obj.field, [])

    def test_set_value(self) -> None:
        """Testing CommaSeparatedValuesField with setting value"""
        obj = TestModel()

        obj.field = ['a', 'b', 'c']
        obj.save()

        obj.refresh_from_db()
        self.assertEqual(obj.field, ['a', 'b', 'c'])



"""HTTP responses for Web APIs."""

from __future__ import annotations

from typing import (Any, Callable, Collection, Dict, Iterator, List, Optional,
                    Sequence, TYPE_CHECKING, Union)

from django.http import HttpResponse
from django.utils.encoding import force_str
from typing_extensions import NotRequired, TypeAlias, TypedDict

from djblets.http.responses import (EventStreamMessage,
                                    EventStreamMessages,
                                    EventStreamHttpResponse)
from djblets.util.http import (get_http_requested_mimetype,
                               get_url_params_except,
                               is_mimetype_a)
from djblets.webapi.encoders import (JSONEncoderAdapter, WebAPIEncoder,
                                     XMLEncoderAdapter,
                                     get_registered_encoders)
from djblets.webapi.errors import INVALID_FORM_DATA


if TYPE_CHECKING:
    from django.db.models import QuerySet
    from django.forms import Form
    from django.http import HttpRequest

    from djblets.webapi.errors import WebAPIError


#: A dictionary of headers to return from the API.
#:
#: Version Added:
#:     3.2
WebAPIResponseHeaders: TypeAlias = Dict[str, str]


class WebAPIResponseLink(TypedDict):
    """Dictionary containing information on a link.

    Version Added:
        3.2
    """

    #: The HTTP method used to access the link.
    #:
    #: Type:
    #:     str
    method: str

    #: The absolute URL that the link points to.
    #:
    #: This may be None in the case of an error generating a link, such as
    #: due to a circular dependency.
    #:
    #: Version Changed:
    #:     4.0:
    #:     This can now be ``None``.
    #:
    #: Type:
    #:     str
    href: Optional[str]

    #: The optional title for the link.
    #:
    #: This may not be present in the dictionary.
    #:
    #: Type:
    #:     str
    title: NotRequired[str]


#: A type alias for a mapping of link names to link information.
#:
#: Version Added:
#:     3.2
WebAPIResponseLinks: TypeAlias = Dict[str, WebAPIResponseLink]


#: A type alias for an API response payload.
#:
#: Version Added:
#:     3.2
WebAPIResponsePayload: TypeAlias = Dict[Any, Any]


class WebAPIEventStreamMessage(EventStreamMessage):
    """A message generated in an API event stream.

    This is used by :py:class:`WebAPIResponseEventStream` to stream data to
    a client.

    A message may contain an ID, application-defined event type, serialized
    payload data (which may be a serialized object payload or may be custom
    text-based data), and a retry time in milliseconds. These are all optional,
    but any empty message will be skipped.

    Version Added:
        4.0
    """

    #: An API response payload to send in the message.
    #:
    #: Type:
    #:     dict
    obj: NotRequired[WebAPIResponsePayload]

    #: The value for the API response payload's ``stat`` key.
    #:
    #: If not provided, this will default to "ok".
    #:
    #: Type:
    #:     str
    stat: NotRequired[str]


#: An iterator for API event stream messages.
#:
#: Version Added:
#:     4.0
WebAPIEventStreamMessages: TypeAlias = Iterator[WebAPIEventStreamMessage]


#: An API event stream that generates messages.
#:
#: This may be an iterator of messages, or a callable that takes an optional
#: :mailheader:`Last-Event-ID` header value and then yields messages.
#:
#: Version Added:
#:     4.0
WebAPIEventStream: TypeAlias = Union[
    WebAPIEventStreamMessages,
    Callable[[Optional[str]], WebAPIEventStreamMessages],
]


class WebAPIResponse(HttpResponse):
    """An API response, formatted for the desired file format."""

    #: The default list of supported mimetypes for the payload.
    #:
    #: Type:
    #:     list of str
    supported_mimetypes: List[str] = [
        'application/json',
        'application/xml',
    ]

    ######################
    # Instance variables #
    ######################

    #: The response payload that will be encoded.
    #:
    #: This will be made up of both default and caller-provided data.
    #:
    #: Type:
    #:     dict
    api_data: WebAPIResponsePayload

    #: A JSONP callback provided in GET requests.
    #:
    #: This will be set if the caller specifies ``?callback=...``. Otherwise,
    #: it will be ``None``.
    #:
    #: Type:
    #:     str
    callback: Optional[str]

    #: Whether :py:attr:`content` has been computed for the final result.
    #:
    #: This is mostly an implementation detail, but is available to outside
    #: callers to know when the content is finalized.
    #:
    #: Type:
    #:     bool
    content_set: bool

    #: The list of encoders that can be used to try to encode the payload.
    #:
    #: Type:
    #:     list of djblets.webapi.encoders.WebAPIEncoder
    encoders: Sequence[WebAPIEncoder]

    #: Keyword arguments to pass to the encoders.
    #:
    #: Type:
    #:     dict
    encoder_kwargs: Dict[str, Any]

    #: The mimetype used for the response.
    #:
    #: Type:
    #:     str
    mimetype: str

    #: The HTTP request from the client.
    #:
    #: Type:
    #:     django.http.HttpRequest
    request: HttpRequest

    def __init__(
        self,
        request: HttpRequest,
        *,
        obj: WebAPIResponsePayload = {},
        stat: str = 'ok',
        api_format: Optional[str] = None,
        status: int = 200,
        headers: WebAPIResponseHeaders = {},
        encoders: Sequence[WebAPIEncoder] = [],
        encoder_kwargs: Dict[str, Any] = {},
        mimetype: Optional[str] = None,
        supported_mimetypes: Optional[List[str]] = None,
    ) -> None:
        """Initialize the response.

        This will attempt to automatically determine the resulting mimetype,
        if ``mimetype`` is not provided.

        If ``api_format`` is provided (as an argument to this function, through
        the ``api_format=`` query string for GET requests, or through a field
        for POST/PUT requests), it will be used to try to guess the mimetype.

        If one cannot be determined through those values, the
        :mailheader:`Accept` HTTP header will be used instead.

        Finally, if no mimetype can be determined, no content will be generated
        and this will automatically be set to a :http:`400`.

        Version Changed:
            3.2:
            All arguments (except for ``request``) must now be provided as
            keywords.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            obj (dict, optional):
                The object payload to serializle.

                Contents depends on the ``encoder``.

            stat (str, optional):
                The value for the API response payload's ``stat`` key.

            api_format (str, optional):
                An explicit format for the result payload, used to determine
                a default for :py:attr:`mimetype`.

            status (int, optional):
                The HTTP status code for the response.

            headers (dict, optional):
                Custom HTTP headers to include in the response.

            encoders (list of djblets.webapi.encoders.WebAPIEncoder, optional):
                The list of encoders available to encode ``obj``.

                If not provided, all registered encoders will be tried. See
                :py:func:`~djblets.webapi.encoders.get_registered_encoders`.

            encoder_kwargs (dict, optional):
                Keyword arguments to pass when instantiating encoders.

            mimetype (str, optional):
                An explicit mimetype to use for the result.

            supported_mimetypes (list of str, optional):
                A list of supported mimetypes for this response.

                This is used when trying to guess a mimetype from the
                :mailheader:`Accept` header.
        """
        mimetype = _normalize_response_mimetype(
            request=request,
            mimetype=mimetype,
            api_format=api_format,
            supported_mimetypes=(supported_mimetypes or
                                 self.supported_mimetypes))

        if not mimetype:
            self.status_code = 400
            self.content_set = True
            return

        super().__init__(content_type=mimetype,
                         status=status)
        self.request = request
        self.callback = request.GET.get('callback', None)
        self.api_data = {'stat': stat}
        self.api_data.update(obj)
        self.content_set = False
        self.mimetype = mimetype
        self.encoders = encoders or get_registered_encoders()
        self.encoder_kwargs = encoder_kwargs

        for header, value in headers.items():
            self[header] = value

        # Prevent IE8 from trying to download some AJAX responses as if they
        # were files.
        self['X-Content-Type-Options'] = 'nosniff'

    # Note: We're overriding HttpResponse.content, which in Django is untyped
    #       and in the django-stubs package is typed with a utility object.
    #       Because of this, we can't easily maintain signature compatibility,
    #       meaning type checkers will yell at us.
    #
    #       Also, while the getter will return a byte string, the setter will
    #       accept anything. Type checkers seem to be confused by this.
    #
    #       This is why we're using `type: ignore` here and below:
    #
    @property                    # type: ignore
    def content(self) -> bytes:  # type: ignore
        """The encoded API response content.

        This is an overridden version of :py:meth:`HttpResponse._get_content()
        <django.http.HttpResponse._get_content>` that generates the resulting
        content when requested, rather than generating it up-front in the
        constructor, allowing data to be updated after construction but before
        it's ready to be sent to the client.

        Type:
            bytes
        """
        class MultiEncoder(WebAPIEncoder):
            def __init__(
                self,
                encoders: Sequence[WebAPIEncoder],
            ) -> None:
                self.encoders = encoders

            def encode(self, *args, **kwargs) -> str:
                for encoder in self.encoders:
                    result = encoder.encode(*args, **kwargs)

                    if result is not None:
                        return result

                return ''

        if not self.content_set:
            adapter: Optional[Union[JSONEncoderAdapter,
                                    XMLEncoderAdapter]] = None
            encoder = MultiEncoder(self.encoders)

            # See the note above about the check for text/plain.
            if (self.mimetype == 'text/plain' or
                is_mimetype_a(self.mimetype, 'application/json')):
                adapter = JSONEncoderAdapter(encoder)
            elif is_mimetype_a(self.mimetype, "application/xml"):
                adapter = XMLEncoderAdapter(encoder)

            assert adapter is not None
            content: str = adapter.encode(self.api_data,
                                          request=self.request,
                                          **self.encoder_kwargs)

            if self.callback is not None:
                content = f'{self.callback}({content});'

            self.content = content  # type: ignore
            self.content_set = True

        return super().content

    @content.setter
    def content(
        self,
        value: Any,
    ) -> None:
        """Set the content on the response.

        Args:
            value (object):
                The value to set.

                This will be converted to a byte string by Django.
        """
        # We need to call this on the class, which is why we're not using
        # `super()`.
        #
        # Note that type checkers would complain here, but this is correct.
        # Internally, `HttpResponse.content` can take various values and will
        # convert them as appropriate. The `django-stubs` module, if used,
        # incorrectly interprets `HttpResponse.content` as being `bytes`.
        HttpResponse.content.fset(self, value)  # type: ignore


class WebAPIResponsePaginated(WebAPIResponse):
    """A response containing a list of results with pagination.

    This accepts the following parameters to the URL:

    * start - The index of the first item (0-based index).
    * max-results - The maximum number of results to return in the request.

    Subclasses can override much of the pagination behavior of this function.
    While the default behavior operates on a queryset and works on indexes
    within that queryset, subclasses can override this to work on any data
    and paginate in any way they see fit.
    """

    ######################
    # Instance variables #
    ######################

    #: The maximum number of results that will be returned.
    #:
    #: Type:
    #:     int
    max_results: int

    #: The query argument key requesting the maximum number of results.
    #:
    #: Type:
    #:     str
    max_results_param: str

    #: The key used for the next results link in the payload.
    #:
    #: Type:
    #:     str
    next_key: str

    #: The key used for the previous results link in the payload.
    #:
    #: Type:
    #:     str
    prev_key: str

    #: The optional queryset used to construct these results.
    #:
    #: Type:
    #:     django.db.models.QuerySet
    queryset: Optional[QuerySet]

    #: The results to return.
    #:
    #: Type:
    #:     collections.abc.Collection
    results: Collection[Any]

    #: The starting offset for results.
    #:
    #: Type:
    #:     int
    start: int

    #: The query argument key requesting the start offset.
    #:
    #: Type:
    #:     str
    start_param: str

    #: The key used for the total results count in the payload.
    #:
    #: Type:
    #:     str
    total_results_key: str

    def __init__(
        self,
        request: HttpRequest,
        *args,
        queryset: Optional[QuerySet] = None,
        results_key: str = 'results',
        prev_key: str = 'prev',
        next_key: str = 'next',
        total_results_key: str = 'total_results',
        start_param: str = 'start',
        max_results_param: str = 'max-results',
        default_start: int = 0,
        default_max_results: int = 25,
        max_results_cap: int = 200,
        serialize_object_func: Optional[Callable[[object], Any]] = None,
        extra_data: Dict[Any, Any] = {},
        **kwargs,
    ) -> None:
        """Initialize the response.

        Version Changed:
            3.2:
            All arguments (except for ``request``) must now be provided as
            keywords.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple):
                Positional arguments to pass to the parent class.

                This is only for backwards-compatibility, and will be removed
                in Djblets 4.0.

            queryset (django.db.models.QuerySet, optional):
                The optional queryset used to construct these results.

            results_key (str, optional):
                The name of the key used to store the resulting list of items.

            prev_key (str, optional):
                The key used for the previous results link in the payload.

            next_key (str, optional):
                The key used for the next results link in the payload.

            total_results_key (str, optional):
                The key used for the total results count in the payload.

            start_param (str, optional):
                The query argument key requesting the start offset.

            max_results_param (str, optional):
                The query argument key requesting the maximum number of
                results.

            default_start (int, optional):
                The default start offset, if not provided in the request
                using ``start_param``.

            default_max_results (int, optional):
                The default maximum number of results, if not provided in
                the request using ``max_results_param``.

            max_results_cap (int, optional):
                The maximum number of results allowed in a response.

                Any user-requested max will be capped to this value.

            serialize_object_func (callable, optional):
                A function to call to serialize a single result.

            extra_data (dict, optional):
                Extra payload data to merge into the resulting payload.

            **kwargs (dict):
                Keyword arguments to pass to the parent class.
        """
        if 'obj' in kwargs:
            raise ValueError(f'obj= cannot be passed to {type(self)}')

        self.request = request
        self.queryset = queryset
        self.prev_key = prev_key
        self.next_key = next_key
        self.start_param = start_param
        self.max_results_param = max_results_param

        self.start = self.normalize_start(
            request.GET.get(start_param, default_start))

        try:
            self.max_results = \
                min(int(request.GET.get(max_results_param,
                                        default_max_results)),
                    max_results_cap)
        except ValueError:
            self.max_results = default_max_results

        self.results = self.get_results()
        self.total_results = self.get_total_results()

        if self.total_results == 0:
            self.results = []
        elif serialize_object_func:
            self.results = [
                serialize_object_func(obj)
                for obj in self.results
            ]
        else:
            self.results = list(self.results)

        data: WebAPIResponsePayload = {
            results_key: self.results,
            'links': {},
        }
        data.update(extra_data)

        links = data['links']
        assert isinstance(links, dict)
        links.update(self.get_links())

        if total_results_key and self.total_results is not None:
            data[total_results_key] = self.total_results

        super().__init__(request,
                         obj=data,
                         *args,
                         **kwargs)

    def normalize_start(
        self,
        start: Union[int, str],
    ) -> int:
        """Normalize the start value.

        By default, this ensures it's an integer no less than 0.
        Subclasses can override this behavior.

        Args:
            start (int or str):
                The requested starting offset.

        Returns:
            int:
            The normalized starting offset.
        """
        try:
            return max(int(start), 0)
        except ValueError:
            return 0

    def has_prev(self) -> bool:
        """Return whether there's a previous set of results.

        Returns:
            bool:
            ``True`` if there's a previous set of results. ``False`` if
            there is not.
        """
        return self.start > 0

    def has_next(self) -> bool:
        """Return whether there's a next set of results.

        Returns:
            bool:
            ``True`` if there's a next set of results. ``False`` if there
            is not.
        """
        return self.start + len(self.results) < self.total_results

    def get_prev_index(self) -> int:
        """Return the previous index to use for ?start=

        Returns:
            int:
            The previous index. This will never be less than 0.
        """
        return max(0, self.start - self.max_results)

    def get_next_index(self) -> int:
        """Return the next index to use for ?start=

        Returns:
            int:
            The next index.
        """
        return self.start + self.max_results

    def get_results(self) -> Collection[Any]:
        """Return the results for this page.

        Returns:
            collections.abc.Collection:
            The collection of results from the queryset.
        """
        if self.queryset is not None:
            return self.queryset[self.start:self.start + self.max_results]

        return []

    def get_total_results(self) -> int:
        """Return the total number of results across all pages.

        Subclasses can return ``None`` to prevent this field from showing up
        in the payload.

        Returns:
            int:
            The number of resulting items.
        """
        if self.queryset is not None:
            return self.queryset.count()

        return 0

    def get_links(self) -> WebAPIResponseLinks:
        """Return all links used in the payload.

        By default, this only includes pagination links. Subclasses can
        provide additional links.

        Returns:
            dict:
            The dictionary mapping link names to link information.

            See :py:class:`WebAPIResponseLinkDict` for the format of the link
            information dictionaries.
        """
        links: WebAPIResponseLinks = {}

        full_path = self.request.build_absolute_uri(self.request.path)

        query_parameters = get_url_params_except(
            self.request.GET, self.start_param, self.max_results_param)

        if query_parameters:
            query_parameters = f'&{query_parameters}'

        if self.has_prev():
            links[self.prev_key] = {
                'method': 'GET',
                'href': self.build_pagination_url(
                    full_path, self.get_prev_index(),
                    self.max_results, query_parameters),
            }

        if self.has_next():
            links[self.next_key] = {
                'method': 'GET',
                'href': self.build_pagination_url(
                    full_path, self.get_next_index(),
                    self.max_results, query_parameters),
            }

        return links

    def build_pagination_url(
        self,
        full_path: str,
        start: int,
        max_results: int,
        query_parameters: str,
    ) -> str:
        """Build a URL to go to the previous or next set of results.

        Args:
            full_path (str):
                The full path to the API endpoint.

            start (int):
                The start offset for the results.


            max_results (int):
                The maximum number of results to return.

            query_parameters (str):
                Additional query parameters to include in the query string.
                This must start with ``&``.
        """
        return (
            f'{full_path}'
            f'?{self.start_param}={start}'
            f'&{self.max_results_param}={max_results}'
            f'{query_parameters}'
        )


class WebAPIResponseError(WebAPIResponse):
    """A general API error response.

    This contains an error code and a human-readable message. Additional
    data can be provided through ``extra_params`` and ``headers``.
    """

    def __init__(
        self,
        request: HttpRequest,
        err: WebAPIError,
        *args,
        extra_params: Dict[Any, Any] = {},
        headers: WebAPIResponseHeaders = {},
        **kwargs,
    ) -> None:
        """Initialize the response.

        Version Changed:
            3.2:
            All arguments (except for ``request``) must now be provided as
            keywords.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            err (djblets.webapi.errors.WebAPIError):
                The error class to use for the response.

            *args (tuple):
                Positional arguments to pass to the parent class.

                This is only for backwards-compatibility, and will be removed
                in Djblets 4.0.

            extra_params (dict, optional):
                Additional data to include in the root of the payload.

                This should not include an ``err`` key.

            headers (dict, optional):
                Custom HTTP headers to include in the response.

            **kwargs (dict):
                Keyword arguments to pass to the parent class.
        """
        for arg in ('obj', 'stat', 'status'):
            if arg in kwargs:
                raise ValueError(f'{arg}= cannot be passed to {type(self)}')

        err_payload = {
            'code': err.code,
            'msg': err.msg,
        }

        if err.error_type:
            err_payload['type'] = err.error_type

            if err.error_subtype:
                err_payload['subtype'] = err.error_subtype

        if err.detail:
            err_payload['detail'] = err.detail

        if err.trace_id:
            err_payload['trace_id'] = err.trace_id

        errdata: WebAPIResponsePayload = {
            'err': err_payload,
        }
        errdata.update(extra_params)

        headers = headers.copy()

        if callable(err.headers):
            headers.update(err.headers(request))
        else:
            headers.update(err.headers)

        super().__init__(request,
                         obj=errdata,
                         stat='fail',
                         status=err.http_status,
                         headers=headers,
                         *args,
                         **kwargs)


class WebAPIResponseFormError(WebAPIResponseError):
    """An error response designed to return all errors from a form."""

    def __init__(
        self,
        request: HttpRequest,
        form: Form,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the response.

        Version Changed:
            3.2:
            All arguments (except for ``request``) must now be provided as
            keywords.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            form (django.forms.Form):
                The form class to represent in the response.

            *args (tuple):
                Positional arguments to pass to the parent class.

                This is only for backwards-compatibility, and will be removed
                in Djblets 4.0.

            **kwargs (dict):
                Keyword arguments to pass to the parent class.
        """
        for arg in ('error', 'extra_params'):
            if arg in kwargs:
                raise ValueError(f'{arg}= cannot be passed to {type(self)}')

        fields: Dict[str, List[str]] = {
            field: [force_str(e) for e in form.errors[field]]
            for field in form.errors
        }

        super().__init__(
            request,
            INVALID_FORM_DATA,
            extra_params={
                'fields': fields,
            },
            *args,
            **kwargs)


class WebAPIResponseEventStream(EventStreamHttpResponse):
    """A Server-Sent Events (SSE) API response.

    This is a specialization of
    :py:class:`~djblets.http.responses.EventStreamHttpResponse` built to
    serialize API responses as part of Server-Sent Events. This enables
    standard API responses to be serialized as results from an event stream,
    useful for streaming progress updates or batched API call results as part
    of an API call.

    By default, each Server-Sent Event will contain API payload data
    containing, at a minimum, a standard ``"stat": "ok"`` field. This can be
    extended by specifying an object payload. The serialized result of the
    payload will be in JSON format by default, but this can be controlled
    through the standard :mailheader:`Accept` header or ``api_format``
    argument.

    The mimetype for standard data payloads (for entries without an ``event``
    type or with ``event: message``, the spec-provided default) will be sent
    in :mailheader:`X-Event-Data-Mimetype`.

    Version Added:
        4.0
    """

    def __init__(
        self,
        event_stream: WebAPIEventStream,
        *,
        request: HttpRequest,
        api_format: Optional[str] = None,
        status: int = 200,
        headers: WebAPIResponseHeaders = {},
        encoders: Sequence[WebAPIEncoder] = [],
        encoder_kwargs: Dict[str, Any] = {},
        message_data_mimetype: Optional[str] = None,
        supported_mimetypes: Optional[List[str]] = None,
    ) -> None:
        """Initialize the API event stream.

        The provided API event stream source will be processed when sending
        content to the client, and not before.

        Args:
            event_stream (callable or generator):
                The event stream as a generator or a callable yielding
                results.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

            api_format (str, optional):
                An explicit format for the result payload, used to determine
                a default for :py:attr:`message_data_mimetype`.

            status (int, optional):
                The HTTP status code to send in the response.

            headers (dict, optional):
                Custom HTTP headers to include in the response.

            encoders (list of djblets.webapi.encoders.WebAPIEncoder, optional):
                The list of encoders available to encode object payloads.

                If not provided, all registered encoders will be tried. See
                :py:func:`~djblets.webapi.encoders.get_registered_encoders`.

            encoder_kwargs (dict, optional):
                Keyword arguments to pass when instantiating encoders.

            message_data_mimetype (str, optional):
                The mimetype for the standard message data payloads.

                If not provided, one will be automatically determined based
                on ``api_format`` and the request mimetype.

            supported_mimetypes (list of str, optional):
                A list of supported mimetypes for this response.

                This is used when trying to guess a mimetype from the
                :mailheader:`Accept` header.
        """
        if not encoders:
            encoders = get_registered_encoders()

        message_data_mimetype = _normalize_response_mimetype(
            request=request,
            mimetype=message_data_mimetype,
            api_format=api_format,
            supported_mimetypes=(supported_mimetypes or
                                 WebAPIResponse.supported_mimetypes))

        def _gen_events(
            last_id: Optional[str],
        ) -> EventStreamMessages:
            nonlocal event_stream

            if callable(event_stream):
                event_stream = event_stream(last_id)

            for event in event_stream:
                if 'data' in event:
                    data = event['data']
                else:
                    data = (
                        WebAPIResponse(request,
                                       obj=event.get('obj', {}),
                                       stat=event.get('stat', 'ok'),
                                       api_format=api_format,
                                       mimetype=message_data_mimetype,
                                       encoders=encoders,
                                       encoder_kwargs=encoder_kwargs)
                        .content
                    )

                yield {
                    'data': data,
                    'event': event.get('event'),
                    'id': event.get('id'),
                    'retry_ms': event.get('retry_ms')
                }

        super().__init__(_gen_events,
                         request=request,
                         status=status)

        if message_data_mimetype:
            self['X-Event-Data-Mimetype'] = message_data_mimetype

        for header, value in headers.items():
            self[header] = value


def _normalize_response_mimetype(
    *,
    request: HttpRequest,
    mimetype: Optional[str],
    api_format: Optional[str],
    supported_mimetypes: Optional[Sequence[str]],
) -> Optional[str]:
    """Normalize a mimetype for an API response.

    If an explicit mimetype is not provided, one will be computed based on:

    1. The ``api_format``, if provided as a parameter or in the request
       (checked in that order).

    2. The value in the :mailheader:`Accept` header (f allowed in
       ``supported_mimetypes``).

    Version Added:
        4.0

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        mimetype (str):
            An explicit mimetype to use.

        api_format (str):
            An explicit API format to use, instead of checking the request.

        supported_mimetypes (list of str):
            A list of supported mimetypse for :mailheader:`Accept` headers.

    Returns:
        str:
        The response mimetype to use.
    """
    if not mimetype:
        if not api_format:
            if request.method == 'GET':
                api_format = request.GET.get('api_format')
            else:
                api_format = request.POST.get('api_format')

        if not api_format:
            mimetype = get_http_requested_mimetype(
                request=request,
                supported_mimetypes=supported_mimetypes)
        elif api_format == 'json':
            mimetype = 'application/json'
        elif api_format == 'xml':
            mimetype = 'application/xml'

    return mimetype



"""Unit tests for djblets.util.functional.

Version Added:
    4.0
"""

from __future__ import annotations

import re

import kgb

from djblets.testing.testcases import TestCase
from djblets.util.functional import iterable_len, lazy_re_compile


class IterableLenTests(TestCase):
    """Unit tests for iterable_len.

    Version Added:
        4.0
    """

    def test_with_len(self) -> None:
        """Testing iterable_len with iterable containing __len__"""
        self.assertEqual(iterable_len([1, 2, 3, 4, 5]), 5)

    def test_with_generator(self) -> None:
        """Testing iterable_len with generator"""
        gen = (
            i
            for i in range(5)
        )
        self.assertEqual(iterable_len(gen), 5)


class LazyReCompileTests(kgb.SpyAgency, TestCase):
    """Unit tests for lazy_re_compile.

    Version Added:
        4.0
    """

    def test_with_bytes(self) -> None:
        """Testing lazy_re_compile with byte string"""
        self.spy_on(re.compile)

        regex = lazy_re_compile(br'x\d+x', re.I)
        self.assertSpyNotCalled(re.compile)

        self.assertIsNotNone(regex.match(b'X123x'))
        self.assertSpyCallCount(re.compile, 1)

        self.assertIsNotNone(regex.match(b'X123x'))
        self.assertSpyCallCount(re.compile, 1)

    def test_with_str(self) -> None:
        """Testing lazy_re_compile with Unicode string"""
        self.spy_on(re.compile)

        regex = lazy_re_compile(r'x\d+x', re.I)
        self.assertSpyNotCalled(re.compile)

        self.assertIsNotNone(regex.match('X123x'))
        self.assertSpyCallCount(re.compile, 1)

        self.assertIsNotNone(regex.match('X123x'))
        self.assertSpyCallCount(re.compile, 1)



"""Unit tests for StaticMediaBuildContext.

Version Added:
    5.0
"""

from __future__ import annotations

from pathlib import Path

from django.test.utils import override_settings

from djblets.extensions.extension import Extension
from djblets.extensions.packaging.static_media import StaticMediaBuildContext
from djblets.testing.testcases import TestCase


class MyExtension(Extension):
    css_bundles = {
        'css-bundle1': {
            'source_filenames': ('file1.css',),
            'output_filename': 'my-css-bundle1.css',
            'extra_context': {
                'abc': 'def',
            },
        },
        'css-bundle2': {
            'source_filenames': [
                'file2.css',
                'file3.css',
            ],
            'compiler_options': {
                'foo': 'bar',
            },
            'variant': 'data-uri',
        },
    }

    js_bundles = {
        'js-bundle1': {
            'source_filenames': ('file1.js',),
            'output_filename': 'my-js-bundle1.js',
            'extra_context': {
                'abc': 'def',
            },
        },
        'js-bundle2': {
            'source_filenames': [
                'file2.js',
                'file3.js',
            ],
            'compiler_options': {
                'foo': 'bar',
            },
        },
    }


class StaticMediaBuildContextTests(TestCase):
    """Unit tests for StaticMediaBuildContext.

    Version Added:
        5.0
    """

    def test_init(self) -> None:
        """Testing StaticMediaBuildContext.__init__"""
        build_context = StaticMediaBuildContext(
            package_id='my-package',
            css_bundles={
                'css-bundle1': {
                    'source_filenames': ('file1.css',),
                    'output_filename': 'my-css-bundle1.css',
                    'extra_context': {
                        'abc': 'def',
                    },
                },
                'css-bundle2': {
                    'source_filenames': [
                        'file2.css',
                        'file3.css',
                    ],
                    'compiler_options': {
                        'foo': 'bar',
                    },
                    'variant': 'data-uri',
                },
            },
            js_bundles={
                'js-bundle1': {
                    'source_filenames': ('file1.js',),
                    'output_filename': 'my-js-bundle1.js',
                    'extra_context': {
                        'abc': 'def',
                    },
                },
                'js-bundle2': {
                    'source_filenames': [
                        'file2.js',
                        'file3.js',
                    ],
                    'compiler_options': {
                        'foo': 'bar',
                    },
                },
            },
            source_root_dir=Path('/path/to/src'),
            static_dir=Path('/path/to/static'))

        self.assertEqual(build_context.package_id, 'my-package')
        self.assertEqual(str(build_context.build_dir), '/path/to/src/build')
        self.assertEqual(str(build_context.source_root_dir), '/path/to/src')
        self.assertEqual(str(build_context.workspaces_dir),
                         '/path/to/src/.npm-workspaces')
        self.assertEqual(str(build_context.node_modules_dir),
                         '/path/to/src/node_modules')
        self.assertEqual(str(build_context.static_dir), '/path/to/static')
        self.assertEqual(build_context.pipeline_css_bundles, {
            'css-bundle1': {
                'source_filenames': ['file1.css'],
                'output_filename': 'my-css-bundle1.css',
                'extra_context': {
                    'abc': 'def',
                },
            },
            'css-bundle2': {
                'source_filenames': [
                    'file2.css',
                    'file3.css',
                ],
                'output_filename': 'css/css-bundle2.min.css',
                'compiler_options': {
                    'foo': 'bar',
                },
                'variant': 'data-uri',
            },
        })
        self.assertEqual(build_context.pipeline_js_bundles, {
            'js-bundle1': {
                'source_filenames': ['file1.js'],
                'output_filename': 'my-js-bundle1.js',
                'extra_context': {
                    'abc': 'def',
                },
            },
            'js-bundle2': {
                'source_filenames': [
                    'file2.js',
                    'file3.js',
                ],
                'output_filename': 'js/js-bundle2.min.js',
                'compiler_options': {
                    'foo': 'bar',
                },
            },
        })

    def test_create_for_extension(self) -> None:
        """Testing StaticMediaBuildContext.create_for_extension"""
        build_context = StaticMediaBuildContext.create_for_extension(
            extension_cls=MyExtension,
            source_root_dir=Path('/path/to/src'))

        self.assertEqual(build_context.package_id, 'MyExtension')
        self.assertEqual(str(build_context.build_dir), '/path/to/src/build')
        self.assertEqual(str(build_context.source_root_dir), '/path/to/src')
        self.assertEqual(str(build_context.workspaces_dir),
                         '/path/to/src/.npm-workspaces')
        self.assertEqual(str(build_context.node_modules_dir),
                         '/path/to/src/node_modules')
        self.assertEqual(build_context.static_dir,
                         Path(__file__).parent / 'static')
        self.assertEqual(build_context.pipeline_css_bundles, {
            'css-bundle1': {
                'source_filenames': ['file1.css'],
                'output_filename': 'my-css-bundle1.css',
                'extra_context': {
                    'abc': 'def',
                },
            },
            'css-bundle2': {
                'source_filenames': [
                    'file2.css',
                    'file3.css',
                ],
                'output_filename': 'css/css-bundle2.min.css',
                'compiler_options': {
                    'foo': 'bar',
                },
                'variant': 'data-uri',
            },
        })
        self.assertEqual(build_context.pipeline_js_bundles, {
            'js-bundle1': {
                'source_filenames': ['file1.js'],
                'output_filename': 'my-js-bundle1.js',
                'extra_context': {
                    'abc': 'def',
                },
            },
            'js-bundle2': {
                'source_filenames': [
                    'file2.js',
                    'file3.js',
                ],
                'output_filename': 'js/js-bundle2.min.js',
                'compiler_options': {
                    'foo': 'bar',
                },
            },
        })

    def test_create_for_extension_with_overrides(self) -> None:
        """Testing StaticMediaBuildContext.create_for_extension with overrides
        """
        build_context = StaticMediaBuildContext.create_for_extension(
            extension_cls=MyExtension,
            source_root_dir=Path('/path/to/src'),
            build_dir=Path('/path/to/build'),
            package_id='my-package')

        self.assertEqual(build_context.package_id, 'my-package')
        self.assertEqual(str(build_context.build_dir), '/path/to/build')
        self.assertEqual(str(build_context.source_root_dir), '/path/to/src')
        self.assertEqual(str(build_context.workspaces_dir),
                         '/path/to/src/.npm-workspaces')
        self.assertEqual(str(build_context.node_modules_dir),
                         '/path/to/src/node_modules')
        self.assertEqual(build_context.static_dir,
                         Path(__file__).parent / 'static')
        self.assertEqual(build_context.pipeline_css_bundles, {
            'css-bundle1': {
                'source_filenames': ['file1.css'],
                'output_filename': 'my-css-bundle1.css',
                'extra_context': {
                    'abc': 'def',
                },
            },
            'css-bundle2': {
                'source_filenames': [
                    'file2.css',
                    'file3.css',
                ],
                'output_filename': 'css/css-bundle2.min.css',
                'compiler_options': {
                    'foo': 'bar',
                },
                'variant': 'data-uri',
            },
        })
        self.assertEqual(build_context.pipeline_js_bundles, {
            'js-bundle1': {
                'source_filenames': ['file1.js'],
                'output_filename': 'my-js-bundle1.js',
                'extra_context': {
                    'abc': 'def',
                },
            },
            'js-bundle2': {
                'source_filenames': [
                    'file2.js',
                    'file3.js',
                ],
                'output_filename': 'js/js-bundle2.min.js',
                'compiler_options': {
                    'foo': 'bar',
                },
            },
        })

    @override_settings(STATICFILES_DIRS=[
        '/path/to/extension1/something',
        '/path/to/extension2/static',
        '/path/to/extension3/static/subdir',
        ('test', '/path/to/extension4/test'),
    ])
    def test_get_lessc_include_path(self) -> None:
        """Testing StaticMediaBuildContext.get_lessc_include_path"""
        build_context = StaticMediaBuildContext.create_for_extension(
            extension_cls=MyExtension,
            source_root_dir=Path('/path/to/src'))

        self.assertEqual(build_context.get_lessc_include_path(), {
            '/path/to/extension1',
            '/path/to/extension2',
            '/path/to/extension3',
            '/path/to/extension3/static',
            '/path/to/extension4',
            '/path/to/src/node_modules',
        })



"""Various utility views."""

from __future__ import annotations

import logging
from enum import Enum
from typing import Dict, Sequence

from django.conf import settings
from django.core.cache import caches
from django.core.cache.backends.locmem import LocMemCache
from django.db import connections
from django.db.utils import OperationalError
from django.http import HttpRequest, HttpResponseForbidden, JsonResponse
from django.http.response import HttpResponseBase
from django.utils.translation import get_language
from django.views.generic import View
from django.views.i18n import JavaScriptCatalog
from typing_extensions import TypedDict

from djblets.cache.backend import cache_memoize, make_cache_key
from djblets.cache.serials import generate_locale_serial
from djblets.util.symbols import UNSET


logger = logging.getLogger(__name__)
locale_serials: Dict[str, int] = {}


def cached_javascript_catalog(
    request: HttpRequest,
    domain: str = 'djangojs',
    packages: Sequence[str] = [],
) -> HttpResponseBase:
    """A cached version of javascript_catalog.

    Args:
        request (django.http.HttpRequest):
            The HTTP request.

        domain (str):
            The translation domain to use when looking up strings.

        packages (list of str):
            The package names to get strings from.

    Returns:
        django.http.HttpResponse:
        The response to send back to the client.
    """
    global locale_serials

    package_str = '_'.join(packages)

    try:
        serial = locale_serials[package_str]
    except KeyError:
        serial = generate_locale_serial(packages)
        locale_serials[package_str] = serial

    return cache_memoize(
        'jsi18n-%s-%s-%s-%d' % (domain, package_str, get_language(), serial),
        lambda: JavaScriptCatalog.as_view(domain=domain,
                                          packages=packages)(request),
        large_data=True,
        compress_large_data=True)


class HealthCheckStatus(str, Enum):
    """A status indicator for health checks.

    Version Added:
        4.0
    """
    # TODO: this can inherit from StrEnum once we're Python 3.11+

    #: The state when a service is running and can be connected to.
    UP = 'UP'

    #: The state when a service is down or otherwise cannot be connected to.
    DOWN = 'DOWN'

    #: The state when it is unknown if a service is up or down.
    UNKNOWN = 'UNKNOWN'

    def __str__(self) -> str:
        """Return a string representation of the status.

        This exists so we can compare a string against an enum. Once we move to
        StrEnum this can be removed.

        Returns:
            str:
            The string status.
        """
        return str.__str__(self)


class HealthCheckResult(TypedDict):
    """Result structure for health checks.

    Version Added:
        4.0
    """

    #: The individual checks that were run.
    checks: Dict[str, HealthCheckStatus]

    #: A mapping of the check name to the error if the service is down.
    errors: Dict[str, str]

    #: The overall health status of the server.
    status: HealthCheckStatus


class HealthCheckView(View):
    """A view for health checks.

    This will check the status of connected database and cache servers, and
    report whether everything can be accessed. It will return either HTTP 200
    or 503, and the payload will be a JSON blob defined by
    :py:class:`HealthCheckResult`.

    Version Added:
        4.0
    """

    def get(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> HttpResponseBase:
        """Perform a health check.

        This will do a health check on whether the database and cache server
        can be used. If both are accessible, this will return an HTTP 200. If
        not, this will return HTTP 500.

        Args:
            request (django.http.HttpRequest):
                The HTTP request.

            *args (tuple, unused):
                Positional arguments, for future expansion.

            **kwargs (dict, unused):
                Keyword arguments, for future expansion.

        Returns:
            django.http.JsonResponse:
            The response to send back to the client.
        """
        allowed_ips = getattr(settings, 'DJBLETS_HEALTHCHECK_IPS',
                              settings.INTERNAL_IPS)

        if request.META.get('REMOTE_ADDR') not in allowed_ips:
            return HttpResponseForbidden()

        result: HealthCheckResult = {
            'checks': {},
            'errors': {},
            'status': HealthCheckStatus.UP,
        }
        status: int = 200
        success: bool = True

        for key in settings.DATABASES.keys():
            result_key = f'database.{key}'

            try:
                db = connections[key]

                with db.cursor():
                    pass

                result['checks'][result_key] = HealthCheckStatus.UP
            except OperationalError as e:
                logger.error('Health check: unable to connect to database '
                             '"%s": %s',
                             key, e,
                             extra={'request': request})

                success = False
                result['checks'][result_key] = HealthCheckStatus.DOWN
                result['errors'][result_key] = str(e)

        cache_key = make_cache_key('djblets-healthcheck')

        for key in settings.CACHES.keys():
            if key == 'forwarded_backend':
                # This is the backing for a Djblets cache forwarding backend
                # (which will be covered under another cache backend key).
                # We can skip this one.
                continue

            result_key = f'cache.{key}'

            try:
                cache = caches[key]

                if isinstance(cache, LocMemCache):
                    # This is a local memory cache. If it doesn't work, the
                    # server is *really* in trouble. We can probably filter
                    # it out, though.
                    continue

                # Check for a key in the cache.
                #
                # We have to do this as two non-atomic operations:
                #
                # 1. Set a key explicitly.
                # 2. Check that something was set.
                #
                # The reason is that some cache backends (pymemcache notably)
                # has mitigation against intermittent outages/bad connections.
                # They can mask outages until a certain number of failures
                # have occurred, and then re-introduce the failed servers
                # after a period of time.
                #
                # These timeframes cause problems with standard service health
                # check behavior, which often employ a Circuit Breaker pattern,
                # requiring a certain number of failures within a certain
                # amount of time before considering a service unhealthy, and
                # then resetting when considering any successful result. This
                # does not play well with the cache backend mitigations.
                #
                # Since the cache backends return a caller-provided default
                # when in outage mitigation mode, we need to check whether the
                # key we just stored can be returned or just disappears. We
                # don't really care about the value itself, just whether we
                # got back a default.
                #
                # Note that this can have a false-positive if the cache server
                # happens to lose our newly-set key right away, but health
                # checks should be able to handle a sporadic outage report.
                cache.set(cache_key, True)

                if cache.get(cache_key, UNSET) is UNSET:
                    raise Exception('Unable to communicate with cache server')

                result['checks'][result_key] = HealthCheckStatus.UP
            except Exception as e:
                logger.error('Health check: unable to connect to cache '
                             '"%s": %s',
                             key, e,
                             extra={'request': request})

                success = False
                result['checks'][result_key] = HealthCheckStatus.DOWN
                result['errors'][result_key] = str(e)

        if not success:
            status = 503
            result['status'] = HealthCheckStatus.DOWN

        return JsonResponse(result, status=status)



"""Mixins for writing Pipeline compilers.

Version Added:
    4.0
"""

import json
import logging
import os
from typing import List

from djblets.cache.backend import cache_memoize


logger = logging.getLogger(__name__)


class SourceMapStaleCheckMixin:
    """Pipeline compiler mixin for using sourcemaps for stale file checks.

    This mixin allows Pipeline compilers to use generated sourcemaps to
    determine which files were used to generate a compiled file, and to check
    each of those to see if the compile file is stale.

    This requires that the tool performing either CSS or JavaScript compilation
    is capable of generating a sourcemap that includes a comprehensive set of
    all files used in its compilation. Some tools (such as LessCSS) generate
    sourcemaps that only contain a subset, and those cannot be used without
    additional post-processing.

    The list of sources from parsing a sourcemap are cached for an hour, and
    will be invalidated any time the timestamp of the sourcemap or compiled
    file changes.

    Version Added:
        4.0
    """

    #: Cache expiration time for sourcemap sources.
    #:
    #: Version Added:
    #:     4.0
    _SOURCES_CACHE_TIME_SECS: int = 60 * 60

    def get_sourcemap_filename(
        self,
        infile: str,
        outfile: str,
    ) -> str:
        """Return the filename for the sourcemap.

        This defaults to adding ``.map`` to the end of the compiled file path.
        Subclasses can change this if they need to look for a file in a
        different location or with a different name.

        Args:
            infile (str):
                The source file path.

            outfile (str):
                The compiled file path.

        Returns:
            str:
            The sourcemap file path.
        """
        return '%s.map' % outfile

    def is_outdated(
        self,
        infile: str,
        outfile: str,
    ) -> bool:
        """Return whether a compiled file is out of date.

        This will check the timestamps of the source file, compiled file,
        sourcemap, and all files referenced by the sourcemap. If the
        source file or any of the sourcemap's source files have changed (or
        are missing), this will trigger recompilation.

        If a sourcemap is missing, this will log an error and will flag the
        file as being out of date.

        Args:
            infile (str):
                The source file path.

            outfile (str):
                The compiled file path.

        Returns:
            bool:
            ``True`` if the compiled file is out of date. ``False`` if it is
            not.
        """
        sourcemap_filename = self.get_sourcemap_filename(infile, outfile)

        if not os.path.exists(sourcemap_filename):
            if os.path.exists(outfile):
                logging.warning(
                    'Could not find source map "%s" for "%s". Please make '
                    'sure you have configured your Pipeline compiler '
                    'settings to emit a source map for this type of file.',
                    sourcemap_filename, outfile)

            return True

        try:
            infile_mtime = os.path.getmtime(infile)
            outfile_mtime = os.path.getmtime(outfile)
            sourcemap_mtime = os.path.getmtime(sourcemap_filename)
        except (FileNotFoundError, OSError):
            return True

        # There may be a discrepancy between the outfile and sourcemap
        # modification times. We'll trust that nothing was hand-modified,
        # and instead just choose the oldest of the two.
        checked_mtime = min(outfile_mtime, sourcemap_mtime)

        if infile_mtime > checked_mtime:
            return True

        # We'll cache the parsing of sources from the sourcemap, keyed off by
        # the file path and the modification time. This way we don't have to
        # load a potentially large file every time we want to check this.
        classname = type(self).__name__
        sourcemap_sources_key = (
            f'{classname}-sourcemap-sources:{sourcemap_filename}:'
            f'{sourcemap_mtime}'
        )

        try:
            sources = cache_memoize(
                sourcemap_sources_key,
                lambda: self._get_sourcemap_sources(sourcemap_filename))
        except IOError as e:
            logger.exception('Error opening source map file "%s": %s',
                             sourcemap_filename, e)
            return True
        except Exception as e:
            logger.exception('Error parsing source map file "%s": %s',
                             sourcemap_filename, e)
            return True

        # Check each source file to determine if the compiled file is stale.
        base_path = os.path.dirname(outfile)

        try:
            for filename in sources:
                filename = os.path.normpath(os.path.join(base_path, filename))

                if os.path.getmtime(filename) > checked_mtime:
                    return True
        except (FileNotFoundError, OSError):
            return True

        return False

    def _get_sourcemap_sources(
        self,
        sourcemap_path: str,
    ) -> List[str]:
        """Return sources from a sourcemap.

        This will open the sourcemap, deserialize it, and return the list of
        sources.

        Results from this are cached.

        Args:
            sourcemap_path (str):
                The path to the sourcemap.

        Returns:
            list of str:
            The sources listed in the sourcemap.
        """
        # Let exceptions bubble up, so we don't cache error results.
        with open(sourcemap_path, 'r') as fp:
            sourcemap = json.load(fp)

        return sourcemap.get('sources', [])



"""Unit tests for djblets.registries."""

from __future__ import annotations

import time
from threading import Lock, Thread
from typing import Any, Callable, Iterable

from kgb import SpyAgency

from djblets.registries.errors import (AlreadyRegisteredError,
                                       ItemLookupError,
                                       RegistrationError)
from djblets.registries.registry import Registry, OrderedRegistry, UNREGISTER
from djblets.registries.signals import registry_populating
from djblets.testing.testcases import TestCase


class Item:
    """An item used for registry tests."""

    ######################
    # Instance variables #
    ######################

    id: int
    name: str
    fake: bool

    def __init__(
        self,
        **attrs,
    ) -> None:
        """Initialize the item.

        Args:
            **attrs (dict):
                Attribute name and value pairs to set on the item.
        """
        self._attrs = set(attr_name for attr_name in attrs)

        for attr_name, attr_value in attrs.items():
            setattr(self, attr_name, attr_value)

    def __repr__(self) -> str:
        attrs = (
            '%s=%r' % (attr_name, getattr(self, attr_name))
            for attr_name in self._attrs
        )
        return '<Item(%s)>' % ', '.join(attrs)

    def __hash__(self) -> int:
        return id(self)

    def __eq__(
        self,
        other: Any,
    ) -> bool:
        return (type(self) is type(other) and
                self._attrs == other._attrs and
                all(getattr(self, attr_name) == getattr(other, attr_name)
                    for attr_name in self._attrs))


class ThreadRegistry(Registry[Item]):
    """A registry used for thread tests.

    Version Added:
        5.0
    """

    lookup_attrs = ('id',)

    ######################
    # Instance variables #
    ######################

    count: int
    sleep_lock: Lock

    def __init__(self, *args, **kwargs) -> None:
        """Initialize the registry.

        Args:
            *args (tuple):
                Positional arguments to pass to the parent.

            **kwargs (tuple):
                Keyword arguments to pass to the parent.
        """
        super().__init__(*args, **kwargs)

        self.count = 0
        self.sleep_lock = Lock()

    def sleep_first_thread(self) -> None:
        """Sleep the first time a thread calls this method."""
        with self.sleep_lock:
            do_sleep = (self.count == 0)
            self.count += 1

        if do_sleep:
            time.sleep(0.1)


class RegistryTests(SpyAgency, TestCase):
    """Tests for djblets.registries.Registry."""

    def test_empty_by_default(self):
        """Testing Registry instances are created empty"""
        r = Registry()

        self.assertEqual(len(r), 0)
        self.assertEqual(set(r), set())

    def test_get_invalid_attribute(self):
        """Testing Registry.get_item with invalid attributes"""
        r = Registry()

        with self.assertRaises(ItemLookupError):
            r.get('foo', 'bar')

    def test_register_item(self) -> None:
        """Testing Registry.register_item"""
        r = Registry()

        self.spy_on(r.on_item_registering)
        self.spy_on(r.on_item_registered)

        items = [1, 2, 3]

        for item in items:
            r.register(item)

        self.assertEqual(set(r), set(items))

        self.assertSpyCallCount(r.on_item_registering, 3)
        self.assertSpyCalledWith(r.on_item_registering, 1)
        self.assertSpyCalledWith(r.on_item_registering, 2)
        self.assertSpyCalledWith(r.on_item_registering, 3)

        self.assertSpyCallCount(r.on_item_registered, 3)
        self.assertSpyCalledWith(r.on_item_registered, 1)
        self.assertSpyCalledWith(r.on_item_registered, 2)
        self.assertSpyCalledWith(r.on_item_registered, 3)

    def test_register_item_with_thread_conflict(self) -> None:
        """Testing Registry.register_item with same item in different threads
        """
        def _thread_main(
            secs: float,
            item: Item,
            expect_fail: bool,
        ) -> None:
            time.sleep(secs)

            if expect_fail:
                with self.assertRaises(AlreadyRegisteredError):
                    registry.register(item)
            else:
                registry.register(item)

        class TestRegistry(ThreadRegistry):
            def on_item_registering(
                self,
                item: Item,
            ) -> None:
                self.sleep_first_thread()

        registry = TestRegistry()
        registry.populate()
        self.spy_on(registry.on_item_registering)
        self.spy_on(registry.on_item_registered)

        self.assertEqual(len(registry), 0)

        item1 = Item(id=0)
        item2 = Item(id=0)

        self._run_threads(
            target=_thread_main,
            threads_args=[
                [0.1, item1, True],
                [0.05, item2, False],
            ])

        self.assertIs(registry.get('id', 0), item2)

        self.assertSpyCallCount(registry.on_item_registering, 2)
        self.assertSpyCallCount(registry.on_item_registered, 1)

    def test_unregister_item(self):
        """Testing Registry.unregister_item"""
        r = Registry()
        items = [1, 2, 3]

        self.spy_on(r.on_item_unregistering)
        self.spy_on(r.on_item_unregistered)

        for item in items:
            r.register(item)

        for item in items:
            r.unregister(item)

        self.assertEqual(set(r), set())

        self.assertSpyCallCount(r.on_item_unregistering, 3)
        self.assertSpyCalledWith(r.on_item_unregistering, 1)
        self.assertSpyCalledWith(r.on_item_unregistering, 2)
        self.assertSpyCalledWith(r.on_item_unregistering, 3)

        self.assertSpyCallCount(r.on_item_unregistered, 3)
        self.assertSpyCalledWith(r.on_item_unregistered, 1)
        self.assertSpyCalledWith(r.on_item_unregistered, 2)
        self.assertSpyCalledWith(r.on_item_unregistered, 3)

    def test_unregister_item_with_thread_conflict(self) -> None:
        """Testing Registry.unregister_item with same item in different
        threads
        """
        def _thread_main(
            secs: float,
            expect_fail: bool,
        ) -> None:
            time.sleep(secs)

            if expect_fail:
                with self.assertRaises(ItemLookupError):
                    registry.unregister(item1)
            else:
                registry.unregister(item1)

        class TestRegistry(ThreadRegistry):
            def on_item_unregistering(
                self,
                item: Item,
            ) -> None:
                self.sleep_first_thread()

        registry = TestRegistry()
        self.spy_on(registry.on_item_unregistering)
        self.spy_on(registry.on_item_unregistered)

        item1 = Item(id=0)
        registry.register(item1)

        self.assertEqual(len(registry), 1)

        self._run_threads(
            target=_thread_main,
            threads_args=[
                [0.1, True],
                [0.05, False],
            ])

        self.assertEqual(len(registry), 0)
        self.assertSpyCallCount(registry.on_item_unregistering, 2)
        self.assertSpyCallCount(registry.on_item_unregistered, 1)

    def test_unregister_removes_attr_lookups(self):
        """Testing Registry.unregister removes lookup entries"""
        class TestRegistry(Registry[Item]):
            lookup_attrs = ('id',)

        r = TestRegistry()
        items = [Item(id=0), Item(id=1), Item(id=2)]

        for item in items:
            r.register(item)

        for item in items:
            r.unregister(item)

            with self.assertRaises(ItemLookupError):
                r.get('id', item.id)

        self.assertEqual(len(r), 0)

    def test_populate(self) -> None:
        """Testing Registry.populate"""
        item1 = Item(id=0)
        item2 = Item(id=1)

        class TestRegistry(Registry[Item]):
            lookup_attrs = ('id',)

            def get_defaults(self):
                yield item1
                yield item2

        self.spy_on(registry_populating.send)

        r = TestRegistry()
        self.spy_on(r.on_populated)
        self.spy_on(r.on_populating)

        r.populate()

        self.assertEqual(r._registry, {
            'id': {
                0: item1,
                1: item2,
            },
        })

        self.assertSpyCalled(r.on_populated)
        self.assertSpyCalled(r.on_populating)

        self.assertIs(r.get('id', item1.id), item1)
        self.assertIs(r.get('id', item2.id), item2)

        self.assertSpyCalledWith(
            registry_populating.send,
            sender=TestRegistry,
            registry=r)

    def test_populate_with_thread_conflict(self) -> None:
        """Testing Registry.populate with multiple threads"""
        def _thread_main(
            secs: float,
        ) -> None:
            time.sleep(secs)
            registry.populate()

            self.assertEqual(list(registry._items), [item1, item2])

        class TestRegistry(ThreadRegistry):
            def get_defaults(self):
                yield item1
                yield item2

            def on_populating(self) -> None:
                self.sleep_first_thread()

        registry = TestRegistry()
        self.spy_on(registry.on_populating)
        self.spy_on(registry.on_populated)

        item1 = Item(id=0)
        item2 = Item(id=1)

        self._run_threads(
            target=_thread_main,
            threads_args=[
                [0.1],
                [0.05],
            ])

        self.assertEqual(list(registry._items), [item1, item2])
        self.assertSpyCallCount(registry.on_populating, 1)
        self.assertSpyCallCount(registry.on_populated, 1)

    def test_population_on_register(self):
        """Testing Registry.register_item triggers population before
        registration
        """
        original_item = Item(id=0)
        duplicate_item = Item(id=0, fake=True)

        class TestRegistry(Registry[Item]):
            lookup_attrs = ('id',)

            def get_defaults(self):
                yield original_item

        self.spy_on(registry_populating.send)

        r = TestRegistry()

        with self.assertRaises(AlreadyRegisteredError):
            r.register(duplicate_item)

        self.assertIs(r.get('id', original_item.id),
                      original_item)
        self.assertSpyCalledWith(
            registry_populating.send,
            sender=TestRegistry,
            registry=r)

    def test_population_on_unregister(self):
        """Testing Registry.unregister_item triggers population before
        unregistration.
        """
        item = Item(id=0)

        class TestRegistry(Registry[Item]):
            lookup_attrs = ('id',)

            def get_defaults(self):
                yield item

        self.spy_on(registry_populating.send)

        r = TestRegistry()
        r.unregister(item)
        self.assertEqual(len(r), 0)

        self.assertSpyCalledWith(
            registry_populating.send,
            sender=TestRegistry,
            registry=r)

    def test_registering_duplicate(self):
        """Testing Registry.register_item with duplicate items"""
        r = Registry()
        r.register(1)

        with self.assertRaises(AlreadyRegisteredError):
            r.register(1)

    def test_registering_duplicate_attributes(self):
        """Testing Registry.register_item with items that have identical
        attributes
        """
        class TestRegistry(Registry[Item]):
            lookup_attrs = ('id',)

        r = TestRegistry()
        r.register(Item(id=0, name='foo'))

        with self.assertRaises(AlreadyRegisteredError):
            r.register(Item(id=0, name='bar'))

    def test_register_missing_attributes(self):
        """Testing Registry.register_item with items that have missing
        attributes
        """
        class TestRegistry(Registry[Item]):
            lookup_attrs = ('id',)

        r = TestRegistry()

        with self.assertRaises(RegistrationError):
            r.register(Item())

    def test_reset(self) -> None:
        """Testing Registry.reset"""
        item1 = Item(id=0)
        item2 = Item(id=1)

        class TestRegistry(Registry[Item]):
            lookup_attrs = ('id',)

            def get_defaults(self):
                yield item1
                yield item2

        self.spy_on(registry_populating.send)

        r = TestRegistry()
        self.spy_on(r.on_resetting)
        self.spy_on(r.on_reset)

        r.populate()
        r.reset()

        self.assertEqual(r._registry, {
            'id': {},
        })

        self.assertSpyCalled(r.on_resetting)
        self.assertSpyCalled(r.on_reset)

    def test_reset_with_thread_conflict(self) -> None:
        """Testing Registry.reset with multiple threads"""
        def _thread_main(
            secs: float,
        ) -> None:
            self.assertEqual(list(registry._items), [item1, item2])

            time.sleep(secs)
            registry.reset()

            self.assertEqual(list(registry._items), [])

        class TestRegistry(ThreadRegistry):
            def get_defaults(self):
                yield item1
                yield item2

            def on_resetting(self) -> None:
                self.sleep_first_thread()

        item1 = Item(id=0)
        item2 = Item(id=1)

        registry = TestRegistry()
        registry.populate()
        self.spy_on(registry.on_resetting)
        self.spy_on(registry.on_reset)

        self._run_threads(
            target=_thread_main,
            threads_args=[
                [0.1],
                [0.05],
            ])

        self.assertEqual(len(registry._items), 0)
        self.assertSpyCallCount(registry.on_resetting, 1)
        self.assertSpyCallCount(registry.on_reset, 1)

    def test_contains(self):
        """Testing Registry.__contains__"""
        r = Registry()
        r.register(1)

        self.assertIn(1, r)
        self.assertNotIn(2, r)

    def test_error_override(self):
        """Testing Registry error formatting strings"""
        class TestRegistry(Registry[int]):
            errors = {
                UNREGISTER: 'The foo "%(item)s" is unregistered.'
            }

        r = TestRegistry()

        with self.assertRaisesMessage(ItemLookupError,
                                      'The foo "1" is unregistered.'):
            r.unregister(1)

    def _run_threads(
        self,
        target: Callable[..., None],
        threads_args: list[list[Any]],
    ) -> None:
        """Run tests with a specified number of threads.

        This will create multiple threads with the same target and different
        arguments, starting them up and then joining them back to the main
        thread.

        Version Added:
            5.0

        Args:
            target (callable):
                The function for each thread to call when started.

            threads_args (list):
                Arguments to provide for each created thread.

                The number of items in this list dictates the number of
                threads.
        """
        threads = [
            Thread(target=target,
                   args=thread_args)
            for thread_args in threads_args
        ]

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()


class OrderedRegistryTests(TestCase):
    """Tests for djblets.registries.registry.OrderedRegistry."""

    class TestRegistry(OrderedRegistry[int]):
        def get_defaults(self) -> Iterable[int]:
            return [1, 2, 3]

    def setUp(self):
        self.registry = OrderedRegistryTests.TestRegistry()

    def test_iteration_order(self):
        """Testing OrderedRegistry iteration order"""
        self.registry.register(4)
        self.assertListEqual(list(self.registry),
                             [1, 2, 3, 4])

    def test_getitem(self):
        """Testing OrderedRegistry.__getitem__"""
        self.assertEqual(self.registry[0], 1)
        self.assertEqual(self.registry[1], 2)
        self.assertEqual(self.registry[2], 3)

    def test_getitem_negative_indices(self):
        """Testing OrderedRegistry.__getitem__ with negative indices"""
        self.assertEqual(self.registry[-1], 3)
        self.assertEqual(self.registry[-2], 2)
        self.assertEqual(self.registry[-3], 1)

    def test_getitem_invalid_index(self):
        """Testing OrderedRegistry.__getitem__ with an invalid index"""
        with self.assertRaises(TypeError):
            self.registry['foo']  # type: ignore

    def test_getitem_out_of_range(self):
        """Testing OrderedRegistry.__getitem__ with an out of range index"""
        with self.assertRaises(IndexError):
            self.registry[1000]



"""Pipeline compiler for bundling modern JavaScript with rollup.js.

Version Added:
    4.0
"""

import os
import re
from pathlib import Path
from typing import List, Optional, Tuple

from pipeline.conf import settings
from pipeline.compilers import SubProcessCompiler

from djblets.pipeline.compilers.mixins import SourceMapStaleCheckMixin


#: The default regular expression used for rollup.js index files.
#:
#: Version Added:
#:     4.0
DEFAULT_FILE_PATTERN = r'index\.(js|es6|es6\.js|ts)$'


_match_regex = re.compile(settings.get('ROLLUP_FILE_PATTERN') or
                          DEFAULT_FILE_PATTERN)


class RollupCompiler(SourceMapStaleCheckMixin, SubProcessCompiler):
    """A Pipeline compiler for interfacing with rollup.js.

    `rollup.js <https://rollupjs.org/>`_ is a module bundler for JavaScript
    that compiles down a JavaScript codebase, linked together through module
    imports (ES6, CommonJS etc.), into a single JavaScript file. It can manage
    compilation with Babel or TypeScript, and supports a wide variety of
    plugins.

    This compiler makes it easy to develop modern JavaScript while still
    taking advantage of the best of Pipeline, such as managing separate
    bundles, loading bundles into Django templates, and triggering automatic
    recompilation on demand if any files (including those imported via modules)
    in the bundle are stale.

    To use this compiler, you will need an ``index`` file in your Pipeline
    JavaScript bundle, which imports any other modules that should be part
    of that compiled rollup.js bundle. An ``index`` file can be in the
    following forms:

    * :file:`index.js`
    * :file:`index.es6`
    * :file:`index.es6.js`
    * :file:`index.ts`

    Custom filename patterns can be specified by setting
    ``settings.PIPELINE['ROLLUP_FILE_PATTERN']`` to a regular expression string
    (defaults to :py:data:`DEFAULT_FILE_PATTERN`).

    Your Pipeline bundle may consist of just this one ``index`` file, or it
    may include other JavaScript as well. Each ``index`` will be compiled as
    its own rollup.js bundle, and its compiled contents included as part of the
    Pipeline bundle.

    Sourcemaps are used to check if any parts of the bundle are stale,
    triggering automatic recompilation when loading pages. Please note that
    any modules that you want included but are not actively being used within
    that rollup.js bundle may be excluded from automatic recompilation
    detection if you have treeshaking enabled (which is enabled by default).

    You will need to set ``settings.PIPELINE['ROLLUP_BINARY']`` to the path
    of :file:`rollup` and then set any command line arguments needed
    (such as a path to your configuration file) in
    ``settings.PIPELINE['ROLLUP_ARGUMENTS']``.

    The arguments should *not* include ``-c`` / ``--config`` to specify the
    Rollup configuration path. This will be computed automatically, in order
    to ensure the right file is used based on whichever source tree may be
    hosting the input file (such as when a project is consuming another
    project's source files and compiling them).

    Version Changed:
        4.1:
        Added support for :file:`rollup.config.mjs` and
        :file:`rollup.config.ts` files.

    Version Added:
        4.0
    """

    output_extension = 'js'

    def match_file(
        self,
        path: str,
    ) -> bool:
        """Return whether this compiler matches a source file.

        This will look for the following filenames by default:

        * :file:`index.js`
        * :file:`index.es6`
        * :file:`index.es6.js`
        * :file:`index.ts`

        To customize this, set ``settings.PIPELINE['ROLLUP_FILE_PATTERN']``
        to a regular expression string.

        Args:
            path (str):
                The path to the source file being considered.

        Returns:
            bool:
            ``True`` if this compiler will invoke rollup.js on this file.
            ``False`` if it will not.
        """
        return _match_regex.match(os.path.basename(path)) is not None

    def compile_file(
        self,
        infile: str,
        outfile: str,
        outdated: bool = False,
        force: bool = False,
    ) -> None:
        """Compile a file using rollup.js.

        This will cause the file and anything it imports in the tree into
        a rollup.js bundle, for inclusion in the Pipeline bundle.

        Args:
            infile (str):
                The source file.

            outfile (str):
                The destination file.

            outdated (bool, optional):
                Whether the destination file is known to be outdated.

            force (bool, optional):
                Whether to force re-running rollup.js on this file.
        """
        if outdated or force:
            # Look for the root of a tree, containing any of a number of
            # build configuration files.
            tree_root, rollup_config_path = \
                self._find_tree_root(os.path.dirname(infile))

            args: List[str] = [
                settings.ROLLUP_BINARY,
            ]

            if rollup_config_path:
                # We found a rollup.config.js. Make sure we tell rollup.js to
                # use it explicitly.
                rollup_config_filename = os.path.basename(rollup_config_path)

                if rollup_config_filename == 'rollup.config.js':
                    args.append('--bundleConfigAsCjs')
                elif rollup_config_filename == 'rollup.config.ts':
                    args += [
                        '--configPlugin',
                        'typescript',
                    ]

                args += [
                    '-c',
                    rollup_config_path,
                ]

            args += [
                settings.ROLLUP_ARGUMENTS,
                '--sourcemap',
                '-i',
                infile,
                '-o',
                outfile,
            ]

            self.execute_command(args, cwd=tree_root)

    def _find_tree_root(
        self,
        start_dir: str,
    ) -> Tuple[Optional[str], Optional[str]]:
        """Return the root of a source tree for an input file.

        This will scan up the tree, looking for a Rollup configuration file:

        * :file:`rollup.config.js`
        * :file:`rollup.config.mjs`
        * :file:`rollup.config.ts`

        Plus another JavaScript configuration:

        * :file:`.babelrc`
        * :file:`babel.config.json`
        * :file:`tsconfig.json`

        This is used to try to find the proper working directory needed to
        successfully compile a file.

        This is important when a project is consuming and building another
        project's static media, so that the consumed project's configuration
        is applied.

        Version Changed:
            4.1:
            Added scanning for :file:`rollup.config.mjs`,
            :file:`rollup.config.ts`, :file:`babel.config.json`, and
            :file:`tsconfig.json` files.

        Args:
            start_dir (str):
                The starting directory for the search.

        Returns:
            str:
            The root of the source tree, or ``None`` if it could not be found.
        """
        path = Path(start_dir)
        root = Path(path.root)

        while path != root:
            for config_filename in ('rollup.config.js',
                                    'rollup.config.mjs',
                                    'rollup.config.ts'):
                rollup_config_path = Path(path / config_filename)

                if rollup_config_path.exists():
                    # This is the ideal result. We found the top-level of the
                    # tree and the rollup.config.js file.
                    return str(path), str(rollup_config_path)

            if (Path(path / '.babelrc').exists() or
                Path(path / 'babel.config.json').exists() or
                Path(path / 'tsconfig.json').exists()):
                # We didn't find rollup.config.js, but we found another
                # JavaScript configuration file. Consider this the top of
                # the tree.
                return str(path), None

            path = path.parent

        return None, None



"""Custom widgets for Django form fields.

This module contains widgets that correspond to fields provided in
:py:mod:`djblets.forms.fields`.
"""

from __future__ import annotations

import copy
from contextlib import contextmanager
from typing import Any, Dict, List, Optional, Tuple

from django.forms import widgets
from django.forms.widgets import HiddenInput
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

from djblets.conditions import ConditionSet
from djblets.conditions.errors import (ConditionChoiceNotFoundError,
                                       ConditionOperatorNotFoundError)


class AmountSelectorWidget(widgets.MultiWidget):
    """A widget for editing an amount and its unit of measurement.

    Version Added:
        3.3
    """

    #: The name of the template used to render the widget.
    template_name = 'djblets_forms/amount_selector_widget.html'

    def __init__(
        self,
        unit_choices: List[Tuple[Optional[int], str]],
        number_attrs: Optional[Dict[str, Any]] = None,
        select_attrs: Optional[Dict[str, Any]] = None,
        attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Initialize the widget.

        Args:
            unit_choices (list of tuple):
                The unit choices for the field.

                This should be a list of 2-tuples with the following entries:

                Tuple:
                    0 (int or None):
                        The conversion factor of the unit to the base unit.
                        The base unit must always have this value set to 1.
                        For the rest of the units, this will be the number
                        that you need to multiply a value in the base unit
                        by in order to convert it to the given unit.

                    1 (str):
                        The name for the unit.

                The list of unit choices should start with the base unit
                and have the rest of the units follow in increasing order
                of magnitude. You may set a conversion factor of ``None`` for
                a unit choice, which will allow you to save the value for this
                widget as ``None`` instead of as an integer amount. The
                ``None`` unit choice should be placed at the end of the list.

            number_attrs (dict, optional):
                Additional HTML element attributes for the NumberInput widget.

            select_attrs (dict, optional):
                Additional HTML element attributes for the Select widget.

            attrs (dict, optional):
                Additional HTML element attributes for the MultiWidget parent.
        """
        self.widgets: Tuple[widgets.NumberInput, widgets.Select] = (
            widgets.NumberInput(attrs=number_attrs),
            widgets.Select(attrs=select_attrs, choices=unit_choices),
        )
        super().__init__(self.widgets, attrs)

    def decompress(
        self,
        value: Optional[int],
    ) -> Tuple[Optional[int], Optional[int]]:
        """Break up the value into an amount and unit tuple.

        This assumes that the value is stored in the base unit, and will
        find the most appropriate unit to display and convert the amount
        to that unit. The most appropriate unit is the largest unit where
        the amount can be converted to a whole number.

        Args:
            value (int or None):
                The stored value.

        Returns:
            tuple:
            A 2-tuple of:

            Tuple:
                0 (int or None):
                    The amount in the unit.

                1 (int or None):
                    The integer representation of the unit.
        """
        if value is None:
            return None, None

        widget_choices = self.widgets[1].choices
        unit_multiplier = 1

        for choice in reversed(widget_choices):
            unit_multiplier = choice[0]

            if unit_multiplier is None:
                continue

            assert isinstance(unit_multiplier, int)

            if (value % unit_multiplier == 0):
                break

        return value // unit_multiplier, unit_multiplier

    def value_from_datadict(
        self,
        data: Dict[str, Any],
        files: Dict[str, Any],
        name: str,
    ) -> Optional[int]:
        """Return a value for the field from a submitted form.

        This serializes the data POSTed for the form into an integer that the
        field can use and validate. This will convert the integer value to the
        base unit.

         Args:
            data (dict):
                The dictionary containing form data.

            files (dict):
                The dictionary containing uploaded files.

            name (str):
                The field name for the value to load.

        Returns:
            int or None:
            The value to save in the field.
        """
        value, unit_multiplier = super().value_from_datadict(data, files, name)

        if unit_multiplier in ('', None):
            return None
        else:
            return int(value) * int(unit_multiplier)


class ConditionsWidget(widgets.Widget):
    """A widget used to request a list of conditions from the user.

    This is used by :js:class:`~djblets.forms.fields.ConditionsField` to allow
    the user to specify a list of possible conditions, composed of choices,
    operators, and values.

    The conditions are displayed as rows of fields. Each row contains a
    ``<select>`` for the condition choice, operator choice, and a
    choice-provided field for a value.

    Additional conditions can be added by the user dynamically.

    Attributes:
        choices (djblets.conditions.choices.ConditionChoices):
            The condition choices for the field.

        mode_widget (django.forms.widgets.RadioSelect):
            The widget for selecting the mode.

        choice_widget (django.forms.widgets.Select):
            The widget for selecting choices. One of these will be rendered
            for every row.

        operator_widget (django.forms.widgets.Select):
            The widget for selecting operators. One of these will be
            rendered for every row.

        choice_kwargs (dict):
            Optional keyword arguments to pass to each
            :py:class:`~djblets.conditions.choices.BaseConditionChoice`
            constructor. This is useful for more advanced conditions that
            need additional data from the form.

            This can be updated dynamically by the form during initialization.
    """

    #: The name of the template used to render the widget.
    template_name = 'djblets_forms/conditions_widget.html'

    def __init__(self, choices, mode_widget, choice_widget, operator_widget,
                 choice_kwargs=None, attrs=None):
        """Initialize the widget.

        Args:
            choices (djblets.conditions.choices.ConditionChoices):
                The condition choices for the field.

            mode_widget (django.forms.widgets.RadioSelect):
                The widget for selecting the mode.

            choice_widget (django.forms.widgets.Select):
                The widget for selecting choices. One of these will be rendered
                for every row.

            operator_widget (django.forms.widgets.Select):
                The widget for selecting operators. One of these will be
                rendered for every row.

            choice_kwargs (dict):
                Optional keyword arguments to pass to each
                :py:class:`~djblets.conditions.choices.BaseConditionChoice`
                constructor. This is useful for more advanced conditions that
                need additional data from the form.

            attrs (dict, optional):
                Additional HTML element attributes for the widget.
        """
        super(ConditionsWidget, self).__init__(attrs=attrs)

        self.choices = choices
        self.mode_widget = mode_widget
        self.choice_widget = choice_widget
        self.operator_widget = operator_widget
        self.choice_kwargs = choice_kwargs
        self.condition_errors = {}

    @property
    def media(self):
        """Media needed for the widget.

        This is used by the form to return all CSS/JavaScript media that the
        page will need in order to render this widget and any widgets used
        in the condition value fields.

        Type:
            django.forms.widgets.Media
        """
        media = (widgets.Media() +
                 self.choice_widget.media +
                 self.operator_widget.media)

        choices = self.choices.get_choices(choice_kwargs=self.choice_kwargs)

        for choice in choices:
            default_value_field = choice.default_value_field

            if callable(default_value_field):
                default_value_field = default_value_field()

            if hasattr(default_value_field, 'widget'):
                media += default_value_field.widget.media

            for operator in choice.get_operators():
                if operator.has_custom_value_field:
                    if (operator.value_field and
                        hasattr(operator.value_field, 'widget')):
                        media += operator.value_field.widget.media

        return media

    def value_from_datadict(self, data, files, name):
        """Return a value for the field from a submitted form.

        This serializes the data POSTed for the form into a format that the
        field can use and validate.

        Args:
            data (django.http.request.QueryDict):
                The dictionary containing form data.

            files (django.http.request.QueryDict):
                The dictionary containing uploaded files.

            name (unicode):
                The field name for the value to load.

        Returns:
            object:
            The value from the form data.
        """
        try:
            last_id = int(data['%s_last_id' % name]) + 1
        except (KeyError, ValueError):
            last_id = 0

        conditions = []

        for i in range(last_id):
            choice_id = data.get('%s_choice[%s]' % (name, i))

            if choice_id is None:
                # There's no choice with this ID. It was probably deleted.
                continue

            operator_id = data.get('%s_operator[%s]' % (name, i))
            value_name = '%s_value[%s]' % (name, i)

            try:
                choice = self.choices.get_choice(
                    choice_id,
                    choice_kwargs=self.choice_kwargs)
                operator = choice.get_operator(operator_id)
                value_field = self._get_value_field(operator)

                if value_field is None:
                    value = None
                else:
                    value = value_field.get_from_form_data(
                        data, files, value_name)
            except (ConditionChoiceNotFoundError,
                    ConditionOperatorNotFoundError):
                value = data.get(value_name)

            conditions.append({
                'choice': choice_id,
                'op': operator_id,
                'value': value,
            })

        return {
            'mode': data.get('%s_mode' % name),
            'conditions': conditions,
        }

    def render(self, name, value, attrs=None, renderer=None):
        """Render the widget to HTML.

        This will serialize all the choices, operators, and existing
        conditions and render an HTML representation, along with setting up
        JavaScript support for configuring the conditions.

        Args:
            name (unicode):
                The base form field name of the widget.

            value (dict):
                The serialized condition set for the field.

            attrs (dict, optional):
                Additional HTML element attributes for the fields.

            renderer (django.forms.renderers.BaseRenderer, optional):
                The form renderer.

        Returns:
            django.utils.safestring.SafeText:
            The rendered HTML for the widget.
        """
        return render_to_string(self.template_name,
                                self.get_context(name, value, attrs))

    def get_context(self, name, value, attrs):
        """Return context for the widget.

        This will serialize all the choices, operators, and existing
        conditions needed to render the widget.

        Args:
            name (unicode):
                The base form field name of the widget.

            value (dict):
                The serialized condition set for the field.

            attrs (dict):
                Additional HTML element attributes for the fields.

        Returns:
            dict:
            The context data for the widget.
        """
        widget_attrs = self.build_attrs(attrs)
        widget_id = widget_attrs.get('id')
        rendered_rows = []
        rows = []

        # Render the mode radio buttons.
        mode_attrs = widget_attrs

        if widget_id:
            mode_attrs = dict(mode_attrs, id='%s_mode' % name)

        rendered_mode = self.mode_widget.render(
            name='%s_mode' % name,
            value=value.get('mode', ConditionSet.MODE_ALL),
            attrs=mode_attrs)

        for i, condition in enumerate(value['conditions']):
            choice_id = condition['choice']
            operator_id = condition['op']
            condition_value = condition.get('value')
            error = None

            # Try to load the necessary choice and operator.
            #
            # Since these may be dynamically registered, we need to take care
            # not to break or lose information at this stage if the choice or
            # operator is missing. We'll just show the raw data. It won't save,
            # but the user will at least get a suitable error.
            try:
                choice = self.choices.get_choice(
                    choice_id,
                    choice_kwargs=self.choice_kwargs)
                operator = choice.get_operator(operator_id)
                error = self.condition_errors.get(i)
            except ConditionChoiceNotFoundError:
                choice = None
                operator = None
                error = _('This choice no longer exists. You will need to '
                          'delete the condition in order to make changes.')
            except ConditionOperatorNotFoundError:
                operator = None
                error = _('This operator no longer exists. You will need to '
                          'delete the condition in order to make changes.')

            valid = (choice is not None and operator is not None)

            if not valid:
                # Set the dropdowns to be non-editable.
                widget_attrs = dict(widget_attrs, disabled='disabled')

            if choice is None:
                choices = ((choice_id, choice_id),)
            else:
                # Set this to empty so that it'll use the default for the
                # field.
                choices = ()

            if operator is None:
                operators = ((operator_id, operator_id),)
            else:
                # Build the operators specific to this choice.
                operators = (
                    (operator.operator_id, operator.name)
                    for operator in choice.operators
                )

            # Render the list of condition choices.
            choice_name = '%s_choice[%s]' % (name, i)
            choice_attrs = widget_attrs

            if widget_id:
                choice_attrs = dict(choice_attrs,
                                    id='%s_choice_%s' % (widget_id, i))

            with self._add_widget_choices(self.choice_widget, choices):
                rendered_choice = self.choice_widget.render(
                    name=choice_name,
                    value=choice_id,
                    attrs=choice_attrs)

            # Render the list of operators.
            operator_name = '%s_operator[%s]' % (name, i)
            operator_attrs = widget_attrs

            if widget_id:
                operator_attrs = dict(operator_attrs,
                                      id='%s_operator_%s' % (widget_id, i))

            with self._add_widget_choices(self.operator_widget, operators):
                rendered_operator = self.operator_widget.render(
                    name=operator_name,
                    value=operator_id,
                    attrs=operator_attrs)

            if valid:
                value_field = self._get_value_field(operator)

                if value_field is not None:
                    condition_value = \
                        value_field.prepare_value_for_widget(condition_value)
            else:
                # Prepare a hidden input with the values we want to retain,
                # as part of the condition. This is needed so that we can
                # disable the form fields without losing that state when
                # submitting the form (since browsers don't send along data
                # from disabled form fields).
                hidden = widgets.HiddenInput()
                rendered_choice += hidden.render(name=choice_name,
                                                 value=choice_id)
                rendered_operator += hidden.render(name=operator_name,
                                                   value=operator_id)

            # Render the value widget.
            value_attrs = widget_attrs

            if widget_id:
                value_attrs = dict(value_attrs,
                                   id='%s_value_%s' % (widget_id, i))

            # Store the information for the template.
            rendered_rows.append({
                'choice': rendered_choice,
                'operator': rendered_operator,
                'error': error,
            })

            row_data = {
                'choiceID': choice_id,
                'operatorID': operator_id,
                'value': condition_value,
                'valid': valid,
            }

            if error:
                row_data['error'] = error

            rows.append(row_data)

        return {
            'field_id': widget_id,
            'field_name': name,
            'rendered_mode': rendered_mode,
            'rendered_rows': rendered_rows,
            'serialized_choices': [
                self._serialize_choice(temp_choice)
                for temp_choice in self.choices.get_choices(
                    choice_kwargs=self.choice_kwargs)
            ],
            'serialized_rows': rows,
        }

    def _serialize_choice(self, choice):
        """Return a serialized choice for the widget.

        This contains information needed by the JavaScript UI to render the
        choice and its related objects. It's not equivalent to the standard
        serialized version of a choice (which is meant for state storage and
        not widget data).

        Args:
            choice (djblets.conditions.choices.BaseConditionChoice):
                The choice to serialize.

        Returns:
            dict:
            The serialized choice.
        """
        return {
            'id': choice.choice_id,
            'name': choice.name,
            'valueField': self._serialize_value_field(
                self._normalize_value_field(choice.default_value_field)),
            'operators': [
                self._serialize_operator(operator)
                for operator in choice.get_operators()
            ]
        }

    def _serialize_operator(self, operator):
        """Return a serialized operator for the widget.

        This contains information needed by the JavaScript UI to render the
        operator and its related objects. It's not equivalent to the standard
        serialized version of an operator (which is meant for state storage and
        not widget data).

        Args:
            operator (djblets.conditions.operators.BaseConditionOperator):
                The operator to serialize.

        Returns:
            dict:
            The serialized operator.
        """
        data = {
            'id': operator.operator_id,
            'name': operator.name,
            'useValue': True,
        }

        value_field = self._get_value_field(operator)

        if value_field is None:
            # This operator doesn't want any kind of value field.
            data['useValue'] = False
        elif operator.has_custom_value_field:
            # This operator has a field that differs from the default for the
            # choice, so specify it.
            data['valueField'] = self._serialize_value_field(value_field)

        return data

    def _serialize_value_field(self, value_field):
        """Return a serialized value field for the widget.

        The serialized contents will be used to create a JavaScript widget
        for the field.

        Args:
            value_field (djblets.conditions.values.BaseConditionValueField):
                The value field to serialize.

        Returns:
            dict:
            Serialized state for the value field.
        """
        if value_field:
            return {
                'model': {
                    'className': value_field.js_model_class,
                    'data': value_field.get_js_model_data(),
                },
                'view': {
                    'className': value_field.js_view_class,
                    'data': value_field.get_js_view_data(),
                },
            }
        else:
            return {}

    def _get_value_field(self, operator):
        """Return the normalized value field for an operator.

        If the operator's value field is a function, the resulting value field
        from calling that function will be returned. Otherwise, the value field
        will be returned directly.

        Args:
            operator (djblets.conditions.operators.BaseConditionOperator):
                The operator containing the value field.

        Returns:
            djblets.conditions.values.BaseConditionValueField:
            The resulting value field.
        """
        return self._normalize_value_field(operator.value_field)

    def _normalize_value_field(self, value_field):
        """Normalize and return a value field.

        If ``value_field`` is a function, the resulting value field from
        calling that function will be returned. Otherwise, ``value_field``
        will be returned directly.

        Args:
            value_field (object):
            A :py:class:`~djblets.conditions.values.BaseConditionValueField`
            or a function returning one.

        Returns:
            djblets.conditions.values.BaseConditionValueField:
            The resulting value field.
        """
        if callable(value_field):
            value_field = value_field()

        return value_field

    @contextmanager
    def _add_widget_choices(self, widget, choices):
        """Temporarily add choices to a widget.

        This temporary appends the provided list of choices to a widget's
        existing list, allowing some additional choices to be there for
        rendering purposes. After the caller's logic is done, the list of
        choices will be reset.

        Args:
            widget (django.forms.fields.widgets.ChoiceInput):
                The choice widget.

            choices (list):
                The list of choices to temporarily append.
        """
        old_choices = widget.choices
        widget.choices = widget.choices + list(choices)

        yield

        widget.choices = old_choices

    def __deepcopy__(self, memo):
        """Return a deep copy of the widget.

        This will return a deep copy of this widget and all subwidgets, so that
        a particular form instance can easily manipulate state without
        affecting other instances.

        Only the state that may be manipulated by an instance will be deep
        copied.

        Args:
            memo (dict):
                The memo dictionary used to track IDs to objects.

        Returns:
            ConditionsWidget:
            A deep copy of this widget's instance.
        """
        obj = super(ConditionsWidget, self).__deepcopy__(memo)
        obj.mode_widget = copy.deepcopy(self.mode_widget, memo)
        obj.choice_widget = copy.deepcopy(self.choice_widget, memo)
        obj.operator_widget = copy.deepcopy(self.operator_widget, memo)
        obj.condition_errors = copy.deepcopy(self.condition_errors, memo)
        obj.choice_kwargs = self.choice_kwargs.copy()

        return obj


class CopyableTextInput(widgets.TextInput):
    """A TextInput widget that renders a link to copy its contents."""

    template_name = 'djblets_forms/copyable_text_input.html'

    def render(self, name, value, attrs=None, renderer=None):
        """Render the widget.

        Args:
            name (unicode):
                The name of the widget.

            value (unicode):
                The value of the widget.

            attrs (dict):
                The attributes of the widget.

            renderer (django.forms.renderers.BaseRenderer, optional):
                The form renderer.

        Returns:
            django.utils.safestring.SafeText:
            The rendered widget.
        """
        context = self.get_context(name, value, attrs)
        field = self._render('django/forms/widgets/text.html', context,
                             renderer)

        return render_to_string(
            self.template_name,
            {
                'field': field,
                'id': attrs['id'],
            })


class ListEditWidget(widgets.Widget):
    """A widget for editing a list of values.

    .. note: The ``djblets-forms`` CSS and JS bundles are required to use this
             widget.
    """

    template_name = 'djblets_forms/list_edit_widget.html'

    def __init__(self, attrs=None, sep=',', value_widget=widgets.TextInput):
        """Initialize the widget.

        Version Changed:
            3.0:
            Added the ``value_widget`` parameter and the ability for this
            widget to handle any list of values instead of only strings.

        Args:
            attrs (dict, optional):
                A dictionary containing HTML attributes to be set
                on the rendered widget.

            sep (str, optional):
                The item separator.

            value_widget (django.forms.widgets.Widget, optional):
                The widget for the values in the list.
        """
        super(ListEditWidget, self).__init__(attrs)

        if isinstance(value_widget, type):
            self._value_widget = value_widget()
        else:
            self._value_widget = value_widget

        self._sep = sep

    @property
    def media(self):
        """Media needed for the widget.

        This is used by the form to return all CSS/JavaScript media that the
        page will need in order to render this widget and any widgets used
        in the condition value fields.

        Version Added:
            3.0

        Type:
            django.forms.widgets.Media
        """
        return widgets.Media() + self._value_widget.media

    def render(self, name, value, attrs=None, renderer=None):
        """Render the widget.

        Version Changed:
            3.0:
            The ``value`` parameter can be a string or a list of any
            type of values.

        Args:
            name (str):
                The field name.

            value (list or str):
                The field value.

            attrs (dict, optional):
                Additional attributes.

            renderer (django.forms.renderers.BaseRenderer, optional):
                The form renderer.

        Returns:
            django.utils.safestring.SafeText:
            The rendered widget.
        """
        return render_to_string(self.template_name,
                                self.get_context(name, value, attrs))

    def get_context(self, name, value, attrs):
        """Return context for the widget.

        This will serialize all the values in the list of values
        using the value_widget.

        Version Added:
            3.0

        Args:

            name (str):
                The base form field name of the widget.

            value (dict):
                The list of values.

            attrs (dict):
                Additional HTML element attributes for the fields.

        Returns:
            dict:
            The context data for the widget.
        """
        value_widget = self._value_widget
        attrs = self.build_attrs(attrs)
        id_ = attrs.pop('id', None)
        rendered_rows = []

        if 'class' in attrs:
            attrs['class'] += ' djblets-c-list-edit-widget__input'
        else:
            attrs['class'] = 'djblets-c-list-edit-widget__input'

        for i, val in enumerate(value):
            if id_ is not None:
                row_attrs = dict(attrs, **{
                    'id': f'{id_}_value_{i}',
                })
            else:
                row_attrs = attrs

            rendered_rows.append(value_widget.render(
                name=f'{name}_value[{i}]',
                value=val,
                attrs=row_attrs))

        # Render the default row, which will be used when adding a new row.
        if id_ is not None:
            row_attrs = dict(attrs, **{
                'id': f'{id_}_value___EDIT_LIST_ROW_ID__',
            })
        else:
            row_attrs = attrs

        rendered_empty_row = value_widget.render(
            name=f'{name}_value[__EDIT_LIST_ROW_INDEX__]',
            value=None,
            attrs=row_attrs)

        # Render the initial row, if we don't have anything to show.
        if rendered_rows:
            rendered_initial_row = ''
        else:
            if id_ is not None:
                row_attrs = dict(attrs, **{
                    'id': f'{id_}_value_0',
                })
            else:
                row_attrs = attrs

            rendered_initial_row = value_widget.render(
                name=f'{name}_value[0]',
                value=None,
                attrs=row_attrs)

        return {
            'name': name,
            'id': id_,
            'remove_text': _('Remove this item.'),
            'rendered_initial_row': mark_safe(rendered_initial_row),
            'rendered_rows': rendered_rows,
            'rendered_empty_row': rendered_empty_row,
        }

    def value_from_datadict(self, data, files, name):
        """Return a value for the field from a submitted form.

        This serializes the data POSTed for the form into a format that the
        field can use and validate.

        Version Added:
            3.0

        Args:
            data (django.http.request.QueryDict):
                The dictionary containing form data.

            files (django.http.request.QueryDict):
                The dictionary containing uploaded files.

            name (str):
                The field name for the value to load.

        Returns:
            list or str:
            The list of values for the field or a string of separated
            values for legacy behavior.
        """
        value_widget = self._value_widget
        values = []

        try:
            num_rows = int(data['%s_num_rows' % name])

            for i in range(num_rows):
                values.append(
                    value_widget.value_from_datadict(
                        data,
                        files,
                        '%s_value[%s]' % (name, i)))
        except (KeyError, ValueError):
            # In this case the widget was called but not rendered, so fetch
            # the data from the field name key in the data dict instead of
            # from the special keys that get set when the widget is rendered.
            values = data[name]

        return values

    def id_for_label(self, id_):
        """Return the main ID to use for this widget.

        This intentionally returns ``None`` since there are multiple fields
        under this widget.

        Args:
            id_ (str):
                The ID of the element.

        Returns:
            str:
            ``None`` so that no ``for=`` attribute is rendered on the label.
        """
        return None


class RelatedObjectWidget(HiddenInput):
    """A base class form widget that lets people select one or more objects.

    This is a base class. Extended classes must define their own render()
    method, to render their own widget with their own data.

    This should be used with relatedObjectSelectorView.es6.js, which extends
    a Backbone view to display data.
    """

    # We inherit from HiddenInput in order for the superclass to render a
    # hidden <input> element, but the Djblets siteconfig field template we use
    # doesn't display labels, help text, errors etc. when ``is_hidden`` is
    # True. Setting it to False still gives us the rendering and data handling
    # we want but renders fieldset fields correctly.
    is_hidden = False

    def __init__(self, multivalued=True):
        """Initialize the RelatedObjectWidget.

        Args:
            local_site_name (unicode, optional):
                The name of the LocalSite where the widget is being rendered.

            multivalued (bool, optional):
                Whether or not the widget should allow selecting multiple
                values.
        """
        super(RelatedObjectWidget, self).__init__()

        self.multivalued = multivalued



"""Database field for storing comma-separated values.

Version Added:
    5.0
"""

from __future__ import annotations

from typing import Optional, TYPE_CHECKING, Union

from django.db import models

if TYPE_CHECKING:
    from django.db.backends.base.base import BaseDatabaseWrapper


class CommaSeparatedValuesField(models.CharField):
    """Database field for storing comma-separated values.

    Version Added:
        5.0
    """

    def from_db_value(
        self,
        value: Optional[str],
        *args,
    ) -> Optional[list[str]]:
        """Convert the value coming from the database.

        Args:
            value (str):
                The value to convert.

            *args (tuple, unused):
                Unused additional keyword arguments.

        Returns:
            list of str:
            The converted value.
        """
        if value is None:
            return value
        else:
            return self.to_python(value)

    def to_python(
        self,
        value: Union[str, list],
    ) -> list[str]:
        """Convert the given value from DB representation to Python.

        Args:
            value (str or list):
                The value to convert.

        Returns:
            list of str:
            The converted value.
        """
        if isinstance(value, list):
            return value
        elif value == '':
            return []
        else:
            return value.split(',')

    def get_prep_value(
        self,
        value: Optional[list[str]],
    ) -> str:
        """Convert the given Python representation to DB format.

        Args:
            value (list of str):
                The value to convert.

        Returns:
            str:
            The data to store in the database.
        """
        if value is None:
            return ''
        else:
            return ','.join(value)

    def get_db_prep_value(
        self,
        value: Union[list[str], str, None],
        connection: BaseDatabaseWrapper,
        prepared: bool = False,
    ) -> str:
        """Return the prepared value for saving to the database.

        Args:
            value (list of str or str):
                The value to save.

            connection (django.db.backends.base.base.BaseDatabaseWrapper):
                The database connection.

            prepared (bool, optional):
                Whether the value has already been prepared.
        """
        if not prepared and not isinstance(value, str):
            return self.get_prep_value(value)
        else:
            assert isinstance(value, str)
            return value

    def value_to_string(
        self,
        obj: models.Model,
    ) -> str:
        """Return the value as a string.

        Args:
            obj (django.db.models.Model):
                The model that this field is a part of.

        Returns:
            str:
            A string representation of the field contents.
        """
        return self.get_prep_value(self.value_from_object(obj))

    def get_default(self) -> list:
        """Return the default value for the field.

        Returns:
            list:
            The default value for the field.
        """
        if self.has_default() and not callable(self.default):
            return self.default

        default = super().get_default()

        if default == '':
            return []

        return default
